# Copyright 2023 StreamSets Inc.

# fmt: off
import pytest

from streamsets.sdk.utils import get_random_string

# fmt: on

NUM_PIPELINES = 12


@pytest.fixture(scope="module")
def sample_pipelines(sch, sch_authoring_sdc_id):
    """A set of trivial pipelines:

    dev_data_generator >> trash
    """
    pipelines = []

    for i in range(NUM_PIPELINES):
        pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

        dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
        trash = pipeline_builder.add_stage('Trash')
        dev_data_generator >> trash

        pipeline_name = 'pipeline_stage_test_{}_{}'.format(i, get_random_string())
        pipeline = pipeline_builder.build(pipeline_name)
        sch.publish_pipeline(pipeline)

        pipelines.append(pipeline)

    try:
        yield pipelines
    finally:
        for pipeline in pipelines:
            sch.delete_pipeline(pipeline)


@pytest.fixture(scope="module")
def sample_pipeline(sch, sch_authoring_sdc_id):
    """A trivial pipelines:

    dev_data_generator >> trash
    """
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
    trash = pipeline_builder.add_stage('Trash')
    dev_data_generator >> trash
    pipeline = pipeline_builder.build('sample_pipeline_{}'.format(get_random_string()))
    sch.publish_pipeline(pipeline)

    try:
        yield pipeline
    finally:
        sch.delete_pipeline(pipeline)


@pytest.fixture(scope="module")
def sample_fragment(sch, sch_authoring_sdc_id):
    pipeline_builder = sch.get_pipeline_builder(
        engine_type='data_collector', engine_id=sch_authoring_sdc_id, fragment=True
    )
    dev = pipeline_builder.add_stage('Dev Data Generator')
    evaluator = pipeline_builder.add_stage('Expression Evaluator')

    dev >> evaluator
    fragment = pipeline_builder.build('sample_pipeline_{}'.format(get_random_string()))

    sch.publish_pipeline(fragment)

    try:
        yield fragment
    finally:
        sch.delete_pipeline(fragment)


@pytest.fixture(scope='module', params=['data_collector', 'transformer', 'snowflake'])
def engine_type(request):
    yield request.param


@pytest.fixture(scope='function')
def pipeline_builder(request, sch, engine_type):
    engine_args = {
        'data_collector': 'sch_authoring_sdc_id',
        'transformer': 'sch_authoring_transformer_id',
        'snowflake': None,
    }
    engine_id_arg = engine_args[engine_type]
    if engine_id_arg:
        extra_args = {'engine_id': request.getfixturevalue(engine_id_arg)}
    else:
        extra_args = {}

    pipeline_builder = sch.get_pipeline_builder(engine_type, **extra_args)

    yield pipeline_builder


# SDC Tests for stage lanes
def test_output_lanes_created_on_instantiation(sch, sch_authoring_sdc_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

    dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
    data_parser = pipeline_builder.add_stage('Data Parser')

    assert len(dev_data_generator.output_lanes) == dev_data_generator.output_streams
    assert len(data_parser.output_lanes) == data_parser.output_streams


def test_output_lanes_not_created_for_0_output_stream_stages(sch, sch_authoring_sdc_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

    trash = pipeline_builder.add_stage('Trash')

    assert len(trash.output_lanes) == 0


def test_output_lanes_not_created_on_stage_connection(sch, sch_authoring_sdc_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

    dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
    data_parser = pipeline_builder.add_stage('Data Parser')
    trash = pipeline_builder.add_stage('Trash')

    assert len(dev_data_generator.output_lanes) == dev_data_generator.output_streams
    assert len(data_parser.output_lanes) == data_parser.output_streams
    assert len(data_parser.input_lanes) == 0
    assert len(trash.input_lanes) == 0

    dev_data_generator_output_lanes = len(dev_data_generator.output_lanes)
    data_parser_output_lanes = len(data_parser.output_lanes)

    dev_data_generator >> data_parser
    data_parser >> trash

    assert len(dev_data_generator.output_lanes) == dev_data_generator_output_lanes
    assert len(data_parser.output_lanes) == data_parser_output_lanes
    assert len(data_parser.input_lanes) == 1
    assert len(trash.input_lanes) == 1


def test_add_event_lane_to_stage_with_no_output_lanes(sch, sch_authoring_sdc_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

    dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
    s3 = pipeline_builder.add_stage('Amazon S3', type='destination')
    trash = pipeline_builder.add_stage('Trash')

    dev_data_generator >> s3
    s3 >= trash

    # asset that s3 has no output_lanes
    assert len(s3.output_lanes) == 0

    # assert that added input_lane was an event lane
    assert len(trash.input_lanes) == 1
    assert s3.instance_name + '_EventLane' == trash.input_lanes[0]

    # publish and make sure the connection still exists
    pipeline = pipeline_builder.build('TEST {}'.format(get_random_string()))

    try:
        sch.publish_pipeline(pipeline)
        pipeline = sch.pipelines.get(name=pipeline.name)
        stages = pipeline.stages

        s3, trash = stages[1], stages[2]

        # assert that event lane still exists
        assert len(s3.output_lanes) == 0
        assert len(trash.input_lanes) == 1
        assert s3.instance_name + '_EventLane' == trash.input_lanes[0]

    except Exception as e:
        raise e

    finally:
        sch.delete_pipeline(pipeline)


def test_input_output_lanes_names_match_appropriate_instance_names(sch, sch_authoring_sdc_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

    dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
    data_parser = pipeline_builder.add_stage('Data Parser')
    trash = pipeline_builder.add_stage('Trash')

    dev_data_generator >> data_parser
    data_parser >> trash
    pipeline = pipeline_builder.build('simple pipeline')

    dev_data_generator, data_parser, trash = pipeline.stages[0], pipeline.stages[1], pipeline.stages[2]

    assert len(dev_data_generator.output_lanes) == 1
    assert len(data_parser.input_lanes) == 1
    assert len(data_parser.output_lanes) == 1
    assert len(trash.input_lanes) == 1

    assert dev_data_generator.output_lanes[0].split('OutputLane')[0] == dev_data_generator.instance_name
    assert data_parser.input_lanes[0].split('OutputLane')[0] == dev_data_generator.instance_name
    assert data_parser.output_lanes[0].split('OutputLane')[0] == data_parser.instance_name
    assert trash.input_lanes[0].split('OutputLane')[0] == data_parser.instance_name


def test_connect_stages_with_multiple_output_streams(sch, sch_authoring_sdc_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

    tail = pipeline_builder.add_stage('File Tail')
    t1 = pipeline_builder.add_stage('Trash')
    t2 = pipeline_builder.add_stage('Trash')
    t3 = pipeline_builder.add_stage('Trash')
    t4 = pipeline_builder.add_stage('Trash')

    tail >> t1
    tail >> t2
    tail >> t3
    tail >> t4
    pipeline = pipeline_builder.build('simple pipeline')

    tail, t1, t2, t3, t4 = (
        pipeline.stages[0],
        pipeline.stages[1],
        pipeline.stages[2],
        pipeline.stages[3],
        pipeline.stages[4],
    )

    # Assert input/output lanes are as expected
    assert len(tail.output_lanes) == 2
    assert len(t1.input_lanes) == 1
    assert len(t2.input_lanes) == 1
    assert len(t3.input_lanes) == 1
    assert len(t4.input_lanes) == 1

    # Assert that the first trash stage was given the first output lane, and all other trash stages were given the
    # second
    assert tail.output_lanes[0] == t1.input_lanes[0]
    assert tail.output_lanes[1] == t2.input_lanes[0]
    assert tail.output_lanes[1] == t3.input_lanes[0]
    assert tail.output_lanes[1] == t4.input_lanes[0]
    assert (
        tail.output_lanes[1] not in t1.input_lanes
        and tail.output_lanes[0] not in t2.input_lanes
        and tail.output_lanes[0] not in t3.input_lanes
        and tail.output_lanes[0] not in t4.input_lanes
    )


def test_connect_multiple_stages_at_once_with_stage_with_multiple_output_streams(sch, sch_authoring_sdc_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

    tail = pipeline_builder.add_stage('File Tail')
    t1 = pipeline_builder.add_stage('Trash')
    t2 = pipeline_builder.add_stage('Trash')
    t3 = pipeline_builder.add_stage('Trash')
    t4 = pipeline_builder.add_stage('Trash')

    tail >> [t1, t2, t3, t4]
    pipeline = pipeline_builder.build('simple pipeline')

    tail, t1, t2, t3, t4 = (
        pipeline.stages[0],
        pipeline.stages[1],
        pipeline.stages[2],
        pipeline.stages[3],
        pipeline.stages[4],
    )

    # Assert input/output lanes are as expected
    assert len(tail.output_lanes) == 2
    assert len(t1.input_lanes) == 1
    assert len(t2.input_lanes) == 1
    assert len(t3.input_lanes) == 1
    assert len(t4.input_lanes) == 1

    # Assert that the all trash stages were assigned the same output lane
    assert tail.output_lanes[0] == t1.input_lanes[0]
    assert tail.output_lanes[0] == t2.input_lanes[0]
    assert tail.output_lanes[0] == t3.input_lanes[0]
    assert tail.output_lanes[0] == t4.input_lanes[0]
    assert (
        tail.output_lanes[1] not in t1.input_lanes
        and tail.output_lanes[1] not in t2.input_lanes
        and tail.output_lanes[1] not in t3.input_lanes
        and tail.output_lanes[1] not in t4.input_lanes
    )


def test_manipulate_output_streams(sch, sch_authoring_sdc_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

    tail = pipeline_builder.add_stage('File Tail')

    tail_output_lanes = tail.output_lanes
    tail.output_lanes = ['DUMMY_OUTPUT_LANE']

    # assert that output_lanes have not been changed
    assert tail.output_lanes != ['DUMMY_OUTPUT_LANE']
    assert tail.output_lanes == tail_output_lanes


def test_disconnect_inputs(sch, sch_authoring_sdc_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

    dev_data_gen = pipeline_builder.add_stage('Dev Data Generator')
    data_parser = pipeline_builder.add_stage('Data Parser')
    jdbc_multitable_consumer = pipeline_builder.add_stage('JDBC Multitable Consumer')
    trash = pipeline_builder.add_stage('Trash')

    dev_data_gen >> trash
    data_parser >> trash
    jdbc_multitable_consumer >> trash

    pipeline = pipeline_builder.build('simple pipeline')

    # assert that the input/output lanes of all stages are accurate
    trash = pipeline.stages[-1]
    assert trash.instance_name == 'Trash_01'
    assert len(trash.input_lanes) == 3

    dev_data_gen, data_parser, jdbc_multitable_consumer = pipeline.stages[0], pipeline.stages[1], pipeline.stages[2]
    assert len(dev_data_gen.output_lanes) == 1
    assert len(data_parser.output_lanes) == 1
    assert len(jdbc_multitable_consumer.output_lanes) == 1

    # assert that only input lanes have been changed appropriately after disconnect_inputs is called
    trash.disconnect_input_lanes([dev_data_gen, data_parser])
    assert len(trash.input_lanes) == 1
    assert len(jdbc_multitable_consumer.output_lanes) == 1
    assert trash.input_lanes[0] == jdbc_multitable_consumer.output_lanes[0]
    assert len(dev_data_gen.output_lanes) == 1
    assert len(data_parser.output_lanes) == 1


def test_disconnect_all_inputs_in_pipeline_builder(sch, sch_authoring_sdc_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

    dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
    data_parser = pipeline_builder.add_stage('Data Parser')
    jdbc_multitable_consumer = pipeline_builder.add_stage('JDBC Multitable Consumer')
    trash = pipeline_builder.add_stage('Trash')

    dev_data_generator >> trash
    data_parser >> trash
    jdbc_multitable_consumer >> trash

    # assert that the input/output lanes of all stages are accurate
    assert len(trash._data['inputLanes']) == 3

    assert len(dev_data_generator._data['outputLanes']) == 1
    assert len(data_parser._data['outputLanes']) == 1
    assert len(jdbc_multitable_consumer._data['outputLanes']) == 1

    # assert that only input lanes have been changed after disconnect_inputs is called with all_stages set to True
    trash.disconnect_input_lanes(all_stages=True)
    assert len(trash._data['inputLanes']) == 0
    assert len(dev_data_generator._data['outputLanes']) == 1
    assert len(data_parser._data['outputLanes']) == 1
    assert len(jdbc_multitable_consumer._data['outputLanes']) == 1


def test_disconnect_all_inputs(sch, sch_authoring_sdc_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

    dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
    data_parser = pipeline_builder.add_stage('Data Parser')
    jdbc_multitable_consumer = pipeline_builder.add_stage('JDBC Multitable Consumer')
    trash = pipeline_builder.add_stage('Trash')

    dev_data_generator >> trash
    data_parser >> trash
    jdbc_multitable_consumer >> trash

    pipeline = pipeline_builder.build('simple pipeline')

    # assert that the input/output lanes of all stages are accurate
    trash = pipeline.stages[-1]
    assert trash.instance_name == 'Trash_01'
    assert len(trash.input_lanes) == 3

    dev_data_gen, data_parser, jdbc_multitable_consumer = pipeline.stages[0], pipeline.stages[1], pipeline.stages[2]
    assert len(dev_data_gen.output_lanes) == 1
    assert len(data_parser.output_lanes) == 1
    assert len(jdbc_multitable_consumer.output_lanes) == 1

    # assert that only input lanes have been changed after disconnect_inputs is called with all_stages set to True
    trash.disconnect_input_lanes(all_stages=True)
    assert len(trash.input_lanes) == 0
    assert len(dev_data_gen.output_lanes) == 1
    assert len(data_parser.output_lanes) == 1
    assert len(jdbc_multitable_consumer.output_lanes) == 1


def test_disconnect_inputs_with_incorrect_stage_type(sch, sch_authoring_sdc_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

    dev_data_gen = pipeline_builder.add_stage('Dev Data Generator')
    data_parser = pipeline_builder.add_stage('Data Parser')
    jdbc_multitable_consumer = pipeline_builder.add_stage('JDBC Multitable Consumer')
    trash = pipeline_builder.add_stage('Trash')

    dev_data_gen >> trash
    data_parser >> trash
    jdbc_multitable_consumer >> trash

    pipeline = pipeline_builder.build('simple pipeline')

    # assert that the input/output lanes of all stages are accurate
    trash = pipeline.stages[-1]
    assert trash.instance_name == 'Trash_01'
    assert len(trash.input_lanes) == 3

    dev_data_gen, data_parser, jdbc_multitable_consumer = pipeline.stages[0], pipeline.stages[1], pipeline.stages[2]
    assert len(dev_data_gen.output_lanes) == 1
    assert len(data_parser.output_lanes) == 1
    assert len(jdbc_multitable_consumer.output_lanes) == 1

    # assert that error is raised
    with pytest.raises(AttributeError):
        trash.disconnect_input_lanes(['STRING TYPE THAT IS CLEARLY AN INVALID TYPE'])


def test_disconnect_inputs_with_stage_that_doesnt_exist(sch, sch_authoring_sdc_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

    dev_data_gen = pipeline_builder.add_stage('Dev Data Generator')
    data_parser = pipeline_builder.add_stage('Data Parser')
    trash = pipeline_builder.add_stage('Trash')

    dev_data_gen >> data_parser
    data_parser >> trash

    pipeline1 = pipeline_builder.build('simple pipeline')

    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    jdbc_multitable_consumer = pipeline_builder.add_stage('JDBC Multitable Consumer')
    pipeline_builder.build('simple pipeline 2')

    # assert that the input/output lanes of all stages are accurate
    dev_data_gen, data_parser, trash = pipeline1.stages[0], pipeline1.stages[1], pipeline1.stages[2]
    assert len(dev_data_gen.output_lanes) == 1
    assert len(data_parser.input_lanes) == 1
    assert len(data_parser.output_lanes) == 1
    assert len(trash.input_lanes) == 1

    # assert that nothing has changed
    dev_data_gen.disconnect_input_lanes([data_parser])
    data_parser.disconnect_input_lanes([trash])
    trash.disconnect_input_lanes([jdbc_multitable_consumer])
    assert len(dev_data_gen.output_lanes) == 1
    assert len(data_parser.input_lanes) == 1
    assert len(data_parser.output_lanes) == 1
    assert len(trash.input_lanes) == 1


def test_disconnect_outputs_in_pipeline_builder(sch, sch_authoring_sdc_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

    dev_data_gen = pipeline_builder.add_stage('Dev Data Generator')
    data_parser = pipeline_builder.add_stage('Data Parser')
    jdbc_multitable_consumer = pipeline_builder.add_stage('JDBC Multitable Consumer')
    trash = pipeline_builder.add_stage('Trash')

    dev_data_gen >> data_parser
    dev_data_gen >> jdbc_multitable_consumer
    dev_data_gen >> trash

    assert len(dev_data_gen._data['outputLanes']) == 1
    assert len(data_parser._data['inputLanes']) == 1
    assert len(jdbc_multitable_consumer._data['inputLanes']) == 1
    assert len(trash._data['inputLanes']) == 1

    # assert that only input lanes have been changed appropriately after disconnect_outputs is called
    dev_data_gen.disconnect_output_lanes([data_parser, jdbc_multitable_consumer])

    assert len(dev_data_gen._data['outputLanes']) == 1
    assert len(trash._data['inputLanes']) == 1

    assert len(data_parser._data['inputLanes']) == 0
    assert len(jdbc_multitable_consumer._data['inputLanes']) == 0


def test_disconnect_outputs(sch, sch_authoring_sdc_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

    dev_data_gen = pipeline_builder.add_stage('Dev Data Generator')
    data_parser = pipeline_builder.add_stage('Data Parser')
    jdbc_multitable_consumer = pipeline_builder.add_stage('JDBC Multitable Consumer')
    trash = pipeline_builder.add_stage('Trash')

    dev_data_gen >> data_parser
    dev_data_gen >> jdbc_multitable_consumer
    dev_data_gen >> trash

    pipeline = pipeline_builder.build('simple pipeline')

    # assert that the input/output lanes of all stages are accurate
    dev_data_gen = pipeline.stages[0]
    assert dev_data_gen.instance_name == 'DevDataGenerator_01'
    assert len(dev_data_gen.output_lanes) == 1

    data_parser, jdbc_multitable_consumer, trash = pipeline.stages[1], pipeline.stages[2], pipeline.stages[3]
    assert len(data_parser.input_lanes) == 1
    assert len(jdbc_multitable_consumer.input_lanes) == 1
    assert len(trash.input_lanes) == 1

    # assert that only input lanes have been changed appropriately after disconnect_outputs is called
    dev_data_gen.disconnect_output_lanes([data_parser, jdbc_multitable_consumer])

    assert len(dev_data_gen.output_lanes) == 1
    assert len(trash.input_lanes) == 1

    assert len(data_parser.input_lanes) == 0
    assert len(jdbc_multitable_consumer.input_lanes) == 0


def test_disconnect_all_outputs(sch, sch_authoring_sdc_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

    dev_data_gen = pipeline_builder.add_stage('Dev Data Generator')
    data_parser = pipeline_builder.add_stage('Data Parser')
    jdbc_multitable_consumer = pipeline_builder.add_stage('JDBC Multitable Consumer')
    trash = pipeline_builder.add_stage('Trash')

    dev_data_gen >> data_parser
    dev_data_gen >> jdbc_multitable_consumer
    dev_data_gen >> trash

    pipeline = pipeline_builder.build('simple pipeline')

    # assert that the input/output lanes of all stages are accurate
    dev_data_gen = pipeline.stages[0]
    assert dev_data_gen.instance_name == 'DevDataGenerator_01'
    assert len(dev_data_gen.output_lanes) == 1

    data_parser, jdbc_multitable_consumer, trash = pipeline.stages[1], pipeline.stages[2], pipeline.stages[3]
    assert len(data_parser.input_lanes) == 1
    assert len(jdbc_multitable_consumer.input_lanes) == 1
    assert len(trash.input_lanes) == 1

    dev_data_gen.disconnect_output_lanes(all_stages=True)

    # assert that only input lanes have been changed appropriately after disconnect_outputs is called
    assert len(dev_data_gen.output_lanes) == 1
    assert len(data_parser.input_lanes) == 0
    assert len(jdbc_multitable_consumer.input_lanes) == 0
    assert len(trash.input_lanes) == 0


def test_disconnect_outputs_with_incorrect_stage_type(sch, sch_authoring_sdc_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

    dev_data_gen = pipeline_builder.add_stage('Dev Data Generator')
    data_parser = pipeline_builder.add_stage('Data Parser')
    jdbc_multitable_consumer = pipeline_builder.add_stage('JDBC Multitable Consumer')
    trash = pipeline_builder.add_stage('Trash')

    dev_data_gen >> data_parser
    dev_data_gen >> jdbc_multitable_consumer
    dev_data_gen >> trash

    pipeline = pipeline_builder.build('simple pipeline')

    # assert that the input/output lanes of all stages are accurate
    dev_data_gen = pipeline.stages[0]
    assert dev_data_gen.instance_name == 'DevDataGenerator_01'
    assert len(dev_data_gen.output_lanes) == 1

    data_parser, jdbc_multitable_consumer, trash = pipeline.stages[1], pipeline.stages[2], pipeline.stages[3]
    assert len(data_parser.input_lanes) == 1
    assert len(jdbc_multitable_consumer.input_lanes) == 1
    assert len(trash.input_lanes) == 1

    # assert that error is raised
    with pytest.raises(AttributeError):
        trash.disconnect_output_lanes(['STRING TYPE THAT IS CLEARLY AN INVALID TYPE'])


def test_disconnect_outputs_with_stage_that_doesnt_exist(sch, sch_authoring_sdc_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

    dev_data_gen = pipeline_builder.add_stage('Dev Data Generator')
    data_parser = pipeline_builder.add_stage('Data Parser')
    trash = pipeline_builder.add_stage('Trash')

    dev_data_gen >> data_parser
    data_parser >> trash

    pipeline1 = pipeline_builder.build('simple pipeline')

    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    jdbc_multitable_consumer = pipeline_builder.add_stage('JDBC Multitable Consumer')
    pipeline_builder.build('simple pipeline 2')

    # assert that the input/output lanes of all stages are accurate
    dev_data_gen, data_parser, trash = pipeline1.stages[0], pipeline1.stages[1], pipeline1.stages[2]
    assert len(dev_data_gen.output_lanes) == 1
    assert len(data_parser.input_lanes) == 1
    assert len(data_parser.output_lanes) == 1
    assert len(trash.input_lanes) == 1

    # assert that nothing has changed
    dev_data_gen.disconnect_output_lanes([trash])
    data_parser.disconnect_output_lanes([jdbc_multitable_consumer])
    trash.disconnect_output_lanes([dev_data_gen])
    assert len(dev_data_gen.output_lanes) == 1
    assert len(data_parser.input_lanes) == 1
    assert len(data_parser.output_lanes) == 1
    assert len(trash.input_lanes) == 1


# SDC Tests for removing stage from Pipeline
def test_remove_stage_from_pipeline_object(sch, sch_authoring_sdc_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

    dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
    data_parser = pipeline_builder.add_stage('Data Parser')
    trash = pipeline_builder.add_stage('Trash')

    dev_data_generator >> data_parser
    data_parser >> trash

    pipeline = pipeline_builder.build('simple pipeline')

    stage_names = {stage.stage_name for stage in pipeline.stages}
    assert len(stage_names) == 3
    assert dev_data_generator.stage_name in stage_names
    assert data_parser.stage_name in stage_names
    assert trash.stage_name in stage_names

    assert len(dev_data_generator.output_lanes) == 1
    assert len(data_parser.input_lanes) == 1
    assert len(data_parser.output_lanes) == 1
    assert len(trash.input_lanes) == 1

    # remove stage
    pipeline.remove_stages(pipeline.stages[1])

    # assert stage has been removed
    stage_names = {stage.stage_name for stage in pipeline.stages}
    assert len(stage_names) == 2
    assert dev_data_generator.stage_name in stage_names
    assert data_parser.stage_name not in stage_names

    assert trash.stage_name in stage_names

    # assert only input_lanes have been removed
    assert len(pipeline.stages[0].output_lanes) == 1
    assert len(pipeline.stages[1].input_lanes) == 0


def test_pass_incorrect_type_into_remove_stages(sch, sch_authoring_sdc_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

    dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
    data_parser = pipeline_builder.add_stage('Data Parser')
    trash = pipeline_builder.add_stage('Trash')

    dev_data_generator >> data_parser
    data_parser >> trash

    pipeline = pipeline_builder.build('simple pipeline')

    with pytest.raises(AttributeError):
        pipeline.remove_stages('STRING THAT SHOULD RESULT IN ERROR')


def test_pass_stage_that_doesnt_exist_into_remove_stages(sch, sch_authoring_sdc_id):
    # Build pipeline1
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

    dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
    data_parser = pipeline_builder.add_stage('Data Parser')
    trash = pipeline_builder.add_stage('Trash')

    dev_data_generator >> data_parser
    data_parser >> trash

    pipeline1 = pipeline_builder.build('simple pipeline 1')

    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    pipeline_builder.add_stage('JDBC Multitable Consumer')
    pipeline_builder.add_stage('Dev Data Generator')
    pipeline2 = pipeline_builder.build('simple pipeline 2')

    with pytest.raises(ValueError):
        pipeline1.remove_stages(pipeline2.stages[0])

    with pytest.raises(ValueError):
        pipeline1.remove_stages(pipeline2.stages[1])

    # Nothing should change within pipeline1
    stage_names = {stage.stage_name for stage in pipeline1.stages}
    assert len(stage_names) == 3
    assert dev_data_generator.stage_name in stage_names
    assert data_parser.stage_name in stage_names
    assert trash.stage_name in stage_names


def test_remove_stage_from_pipeline_builder(sch, sch_authoring_sdc_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

    dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
    data_parser = pipeline_builder.add_stage('Data Parser')
    trash = pipeline_builder.add_stage('Trash')

    dev_data_generator >> data_parser
    data_parser >> trash

    stage_names = {stage['stageName'] for stage in pipeline_builder._pipeline[pipeline_builder._config_key]['stages']}
    assert len(stage_names) == 3
    assert dev_data_generator.stage_name in stage_names
    assert data_parser.stage_name in stage_names
    assert trash.stage_name in stage_names

    # remove stage
    pipeline_builder.remove_stage(data_parser)

    stage_names = {stage['stageName'] for stage in pipeline_builder._pipeline[pipeline_builder._config_key]['stages']}
    assert len(stage_names) == 2
    assert dev_data_generator.stage_name in stage_names
    assert data_parser.stage_name not in stage_names
    assert trash.stage_name in stage_names


def test_remove_connected_stages_from_pipeline_builder(sch, sch_authoring_sdc_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

    dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
    data_parser = pipeline_builder.add_stage('Data Parser')
    trash = pipeline_builder.add_stage('Trash')

    # connect stages
    dev_data_generator >> data_parser
    data_parser >> trash

    stage_names = {stage['stageName'] for stage in pipeline_builder._pipeline[pipeline_builder._config_key]['stages']}
    assert len(stage_names) == 3
    assert dev_data_generator.stage_name in stage_names
    assert data_parser.stage_name in stage_names
    assert trash.stage_name in stage_names
    assert len(dev_data_generator.output_lanes) == 1
    assert len(data_parser.input_lanes) == 1
    assert len(data_parser.output_lanes) == 1
    assert len(trash.input_lanes) == 1

    # remove stage
    pipeline_builder.remove_stage(data_parser)

    stage_names = {stage['stageName'] for stage in pipeline_builder._pipeline[pipeline_builder._config_key]['stages']}
    assert len(stage_names) == 2
    assert dev_data_generator.stage_name in stage_names
    assert data_parser.stage_name not in stage_names
    assert trash.stage_name in stage_names

    # assert only input_lanes have been removed
    assert len(dev_data_generator.output_lanes) == 1
    assert len(trash.input_lanes) == 0


def test_remove_stage_from_pipeline_builder_that_doesnt_exist(sch, sch_authoring_sdc_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

    dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
    data_parser = pipeline_builder.add_stage('Data Parser')
    trash = pipeline_builder.add_stage('Trash')

    stage_names = {stage['stageName'] for stage in pipeline_builder._pipeline[pipeline_builder._config_key]['stages']}
    assert len(stage_names) == 3
    assert dev_data_generator.stage_name in stage_names
    assert data_parser.stage_name in stage_names
    assert trash.stage_name in stage_names

    # remove stage
    with pytest.raises(Exception):
        pipeline_builder.remove_stages('STAGE THAT SHOULD RESULT IN ERROR')


# Transformer Tests for stage lanes
def test_input_output_lanes_property_values_transformer(sch, sch_authoring_transformer_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='transformer', engine_id=sch_authoring_transformer_id)

    aggregate = pipeline_builder.add_stage('Aggregate')
    join = pipeline_builder.add_stage('Join')
    trash = pipeline_builder.add_stage('Trash')

    aggregate >> join
    join >> trash

    pipeline = pipeline_builder.build('simple pipeline')

    aggregate, join, trash = pipeline.stages[0], pipeline.stages[1], pipeline.stages[2]

    assert len(aggregate.output_lanes) == 1
    assert len(join.input_lanes) == 1
    assert len(join.output_lanes) == 1
    assert len(trash.input_lanes) == 1

    assert aggregate.output_lanes[0].split('OutputLane')[0] == aggregate.instance_name
    assert join.input_lanes[0].split('OutputLane')[0] == aggregate.instance_name
    assert join.output_lanes[0].split('OutputLane')[0] == join.instance_name
    assert trash.input_lanes[0].split('OutputLane')[0] == join.instance_name


def test_disconnect_inputs_transformer(sch, sch_authoring_transformer_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='transformer', engine_id=sch_authoring_transformer_id)

    aggregate = pipeline_builder.add_stage('Aggregate')
    join = pipeline_builder.add_stage('Join')
    dev_random = pipeline_builder.add_stage('Dev Random')
    trash = pipeline_builder.add_stage('Trash')

    aggregate >> trash
    join >> trash
    dev_random >> trash

    pipeline = pipeline_builder.build('simple pipeline')

    # assert that the input/output lanes of all stages are accurate
    trash = pipeline.stages[-1]
    assert trash.instance_name == 'Trash_01'
    assert len(trash.input_lanes) == 3

    aggregate, join, dev_random = pipeline.stages[0], pipeline.stages[1], pipeline.stages[2]
    assert len(aggregate.output_lanes) == 1
    assert len(join.output_lanes) == 1
    assert len(dev_random.output_lanes) == 1

    # assert that the input/output lanes have been changed appropriately after disconnect_inputs is called
    trash.disconnect_input_lanes([aggregate, join])
    assert len(trash.input_lanes) == 1
    assert len(dev_random.output_lanes) == 1
    assert trash.input_lanes[0] == dev_random.output_lanes[0]
    assert len(aggregate.output_lanes) == 1
    assert len(join.output_lanes) == 1


def test_disconnect_all_inputs_transformer(sch, sch_authoring_transformer_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='transformer', engine_id=sch_authoring_transformer_id)

    aggregate = pipeline_builder.add_stage('Aggregate')
    join = pipeline_builder.add_stage('Join')
    dev_random = pipeline_builder.add_stage('Dev Random')
    trash = pipeline_builder.add_stage('Trash')

    aggregate >> trash
    join >> trash
    dev_random >> trash

    pipeline = pipeline_builder.build('simple pipeline')

    # assert that the input/output lanes of all stages are accurate
    trash = pipeline.stages[-1]
    assert trash.instance_name == 'Trash_01'
    assert len(trash.input_lanes) == 3

    aggregate, join, dev_random = pipeline.stages[0], pipeline.stages[1], pipeline.stages[2]
    assert len(aggregate.output_lanes) == 1
    assert len(join.output_lanes) == 1
    assert len(dev_random.output_lanes) == 1

    # assert that the input/output lanes have been changed after disconnect_inputs is called with all_stages set to True
    trash.disconnect_input_lanes(all_stages=True)
    assert len(trash.input_lanes) == 0
    assert len(aggregate.output_lanes) == 1
    assert len(join.output_lanes) == 1
    assert len(dev_random.output_lanes) == 1


def test_disconnect_inputs_with_incorrect_stage_type_transformer(sch, sch_authoring_transformer_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='transformer', engine_id=sch_authoring_transformer_id)

    aggregate = pipeline_builder.add_stage('Aggregate')
    join = pipeline_builder.add_stage('Join')
    dev_random = pipeline_builder.add_stage('Dev Random')
    trash = pipeline_builder.add_stage('Trash')

    aggregate >> trash
    join >> trash
    dev_random >> trash

    pipeline = pipeline_builder.build('simple pipeline')

    # assert that the input/output lanes of all stages are accurate
    trash = pipeline.stages[-1]
    assert trash.instance_name == 'Trash_01'
    assert len(trash.input_lanes) == 3

    aggregate, join, dev_random = pipeline.stages[0], pipeline.stages[1], pipeline.stages[2]
    assert len(aggregate.output_lanes) == 1
    assert len(join.output_lanes) == 1
    assert len(dev_random.output_lanes) == 1

    # assert that error is raised
    with pytest.raises(AttributeError):
        trash.disconnect_input_lanes(['STRING TYPE THAT IS CLEARLY AN INVALID TYPE'])


def test_disconnect_inputs_with_stage_that_doesnt_exist_transformer(sch, sch_authoring_transformer_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='transformer', engine_id=sch_authoring_transformer_id)

    aggregate = pipeline_builder.add_stage('Aggregate')
    join = pipeline_builder.add_stage('Join')
    trash = pipeline_builder.add_stage('Trash')

    aggregate >> join
    join >> trash

    pipeline1 = pipeline_builder.build('simple pipeline')

    pipeline_builder = sch.get_pipeline_builder(engine_type='transformer', engine_id=sch_authoring_transformer_id)
    dev_random = pipeline_builder.add_stage('Dev Random')
    pipeline_builder.build('simple pipeline 2')

    # assert that the input/output lanes of all stages are accurate
    aggregate, join, trash = pipeline1.stages[0], pipeline1.stages[1], pipeline1.stages[2]
    assert len(aggregate.output_lanes) == 1
    assert len(join.input_lanes) == 1
    assert len(join.output_lanes) == 1
    assert len(trash.input_lanes) == 1

    # assert that nothing has changed
    aggregate.disconnect_input_lanes([join])
    join.disconnect_input_lanes([trash])
    trash.disconnect_input_lanes([dev_random])
    assert len(aggregate.output_lanes) == 1
    assert len(join.input_lanes) == 1
    assert len(join.output_lanes) == 1
    assert len(trash.input_lanes) == 1


def test_disconnect_outputs_transformer(sch, sch_authoring_transformer_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='transformer', engine_id=sch_authoring_transformer_id)

    aggregate = pipeline_builder.add_stage('Aggregate')
    join = pipeline_builder.add_stage('Join')
    dev_random = pipeline_builder.add_stage('Dev Random')
    trash = pipeline_builder.add_stage('Trash')

    aggregate >> join
    aggregate >> dev_random
    aggregate >> trash

    pipeline = pipeline_builder.build('simple pipeline')

    # assert that the input/output lanes of all stages are accurate
    aggregate = pipeline.stages[0]
    assert len(aggregate.output_lanes) == 1

    join, dev_random, trash = pipeline.stages[1], pipeline.stages[2], pipeline.stages[3]
    assert len(join.input_lanes) == 1
    assert len(dev_random.input_lanes) == 1
    assert len(trash.input_lanes) == 1

    # assert that the input lanes have been changed appropriately after disconnect_outputs is called
    aggregate.disconnect_output_lanes([join, dev_random])

    assert len(aggregate.output_lanes) == 1
    assert len(trash.input_lanes) == 1

    assert len(join.input_lanes) == 0
    assert len(dev_random.input_lanes) == 0


def test_disconnect_all_outputs_transformer(sch, sch_authoring_transformer_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='transformer', engine_id=sch_authoring_transformer_id)

    aggregate = pipeline_builder.add_stage('Aggregate')
    join = pipeline_builder.add_stage('Join')
    dev_random = pipeline_builder.add_stage('Dev Random')
    trash = pipeline_builder.add_stage('Trash')

    aggregate >> join
    aggregate >> dev_random
    aggregate >> trash

    pipeline = pipeline_builder.build('simple pipeline')

    # assert that the input/output lanes of all stages are accurate
    aggregate = pipeline.stages[0]
    assert len(aggregate.output_lanes) == 1

    join, dev_random, trash = pipeline.stages[1], pipeline.stages[2], pipeline.stages[3]
    assert len(join.input_lanes) == 1
    assert len(dev_random.input_lanes) == 1
    assert len(trash.input_lanes) == 1

    aggregate.disconnect_output_lanes(all_stages=True)
    assert len(aggregate.output_lanes) == 1
    assert len(join.input_lanes) == 0
    assert len(dev_random.input_lanes) == 0
    assert len(trash.input_lanes) == 0


def test_disconnect_outputs_with_incorrect_stage_type_transformer(sch, sch_authoring_transformer_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='transformer', engine_id=sch_authoring_transformer_id)

    aggregate = pipeline_builder.add_stage('Aggregate')
    join = pipeline_builder.add_stage('Join')
    dev_random = pipeline_builder.add_stage('Dev Random')
    trash = pipeline_builder.add_stage('Trash')

    aggregate >> join
    aggregate >> dev_random
    aggregate >> trash

    pipeline = pipeline_builder.build('simple pipeline')

    # assert that the input/output lanes of all stages are accurate
    aggregate = pipeline.stages[0]
    assert len(aggregate.output_lanes) == 1

    join, dev_random, trash = pipeline.stages[1], pipeline.stages[2], pipeline.stages[3]
    assert len(join.input_lanes) == 1
    assert len(dev_random.input_lanes) == 1
    assert len(trash.input_lanes) == 1

    # assert that error is raised
    with pytest.raises(AttributeError):
        trash.disconnect_output_lanes(['STRING TYPE THAT IS CLEARLY AN INVALID TYPE'])


def test_disconnect_outputs_with_stage_that_doesnt_exist_transformer(sch, sch_authoring_transformer_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='transformer', engine_id=sch_authoring_transformer_id)

    aggregate = pipeline_builder.add_stage('Aggregate')
    join = pipeline_builder.add_stage('Join')
    trash = pipeline_builder.add_stage('Trash')

    aggregate >> join
    join >> trash

    pipeline1 = pipeline_builder.build('simple pipeline')

    pipeline_builder = sch.get_pipeline_builder(engine_type='transformer', engine_id=sch_authoring_transformer_id)
    dev_random = pipeline_builder.add_stage('Dev Random')
    pipeline_builder.build('simple pipeline 2')

    # assert that the input/output lanes of all stages are accurate
    aggregate, join, trash = pipeline1.stages[0], pipeline1.stages[1], pipeline1.stages[2]
    assert len(aggregate.output_lanes) == 1
    assert len(join.input_lanes) == 1
    assert len(join.output_lanes) == 1
    assert len(trash.input_lanes) == 1

    # assert that nothing has changed
    aggregate.disconnect_output_lanes([trash])
    join.disconnect_output_lanes([dev_random])
    trash.disconnect_output_lanes([aggregate])
    assert len(aggregate.output_lanes) == 1
    assert len(join.input_lanes) == 1
    assert len(join.output_lanes) == 1
    assert len(trash.input_lanes) == 1


# Transformer Tests for removing stage from Pipeline
def test_remove_stage_from_pipeline_object_transformer(sch, sch_authoring_transformer_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='transformer', engine_id=sch_authoring_transformer_id)

    aggregate = pipeline_builder.add_stage('Aggregate')
    join = pipeline_builder.add_stage('Join')
    trash = pipeline_builder.add_stage('Trash')

    aggregate >> join
    join >> trash

    pipeline = pipeline_builder.build('simple pipeline')

    stage_names = {stage.stage_name for stage in pipeline.stages}
    assert len(stage_names) == 3
    assert aggregate.stage_name in stage_names
    assert join.stage_name in stage_names
    assert trash.stage_name in stage_names

    assert len(aggregate.output_lanes) == 1
    assert len(join.input_lanes) == 1
    assert len(join.output_lanes) == 1
    assert len(trash.input_lanes) == 1

    # remove stage
    pipeline.remove_stages(pipeline.stages[1])

    # assert stage has been removed
    stage_names = {stage.stage_name for stage in pipeline.stages}
    assert len(stage_names) == 2
    assert aggregate.stage_name in stage_names
    assert join.stage_name not in stage_names
    assert trash.stage_name in stage_names

    # assert lanes have been removed
    assert len(pipeline.stages[0].output_lanes) == 1
    assert len(pipeline.stages[1].input_lanes) == 0


def test_pass_incorrect_type_into_remove_stages_transformer(sch, sch_authoring_transformer_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='transformer', engine_id=sch_authoring_transformer_id)

    aggregate = pipeline_builder.add_stage('Aggregate')
    join = pipeline_builder.add_stage('Join')
    trash = pipeline_builder.add_stage('Trash')

    aggregate >> join
    join >> trash

    pipeline = pipeline_builder.build('simple pipeline')

    with pytest.raises(AttributeError):
        pipeline.remove_stages('STRING THAT SHOULD RESULT IN ERROR')


def test_pass_stage_that_doesnt_exist_into_remove_stages_transformer(sch, sch_authoring_transformer_id):
    # Build pipeline1
    pipeline_builder = sch.get_pipeline_builder(engine_type='transformer', engine_id=sch_authoring_transformer_id)

    aggregate = pipeline_builder.add_stage('Aggregate')
    join = pipeline_builder.add_stage('Join')
    trash = pipeline_builder.add_stage('Trash')

    aggregate >> join
    join >> trash

    pipeline1 = pipeline_builder.build('simple pipeline 1')

    pipeline_builder = sch.get_pipeline_builder(engine_type='transformer', engine_id=sch_authoring_transformer_id)
    pipeline_builder.add_stage('Dev Random')
    pipeline_builder.add_stage('Join')
    pipeline2 = pipeline_builder.build('simple pipeline 2')

    with pytest.raises(ValueError):
        pipeline1.remove_stages(pipeline2.stages[0])

    with pytest.raises(ValueError):
        pipeline1.remove_stages(pipeline2.stages[1])

    # Nothing should change within pipeline1
    stage_names = {stage.stage_name for stage in pipeline1.stages}
    assert len(stage_names) == 3
    assert aggregate.stage_name in stage_names
    assert join.stage_name in stage_names
    assert trash.stage_name in stage_names


def test_remove_stage_from_pipeline_builder_transformer(sch, sch_authoring_transformer_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='transformer', engine_id=sch_authoring_transformer_id)

    aggregate = pipeline_builder.add_stage('Aggregate')
    join = pipeline_builder.add_stage('Join')
    trash = pipeline_builder.add_stage('Trash')

    stage_names = {stage['stageName'] for stage in pipeline_builder._pipeline[pipeline_builder._config_key]['stages']}
    assert len(stage_names) == 3
    assert aggregate.stage_name in stage_names
    assert join.stage_name in stage_names
    assert trash.stage_name in stage_names

    # remove stage
    pipeline_builder.remove_stage(join)

    stage_names = {stage['stageName'] for stage in pipeline_builder._pipeline[pipeline_builder._config_key]['stages']}
    assert len(stage_names) == 2
    assert aggregate.stage_name in stage_names
    assert join.stage_name not in stage_names
    assert trash.stage_name in stage_names


def test_remove_stage_from_pipeline_builder_that_doesnt_exist_transformer(sch, sch_authoring_transformer_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='transformer', engine_id=sch_authoring_transformer_id)

    aggregate = pipeline_builder.add_stage('Aggregate')
    join = pipeline_builder.add_stage('Join')
    trash = pipeline_builder.add_stage('Trash')

    stage_names = {stage['stageName'] for stage in pipeline_builder._pipeline[pipeline_builder._config_key]['stages']}
    assert len(stage_names) == 3
    assert aggregate.stage_name in stage_names
    assert join.stage_name in stage_names
    assert trash.stage_name in stage_names

    # remove stage
    with pytest.raises(Exception):
        pipeline_builder.remove_stages('STAGE THAT SHOULD RESULT IN ERROR')


def test_connect_outputs_to_multi_lane_fixed_stage(sch, sch_authoring_sdc_id):
    # Building a new pipeline since we need to test the construction of one as part of this test case
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    dev_raw_data_source = pipeline_builder.add_stage('Dev Raw Data Source').set_attributes(
        data_format='JSON', raw_data='12345', stop_after_first_batch=True
    )
    deduplicator = pipeline_builder.add_stage('Record Deduplicator')
    trash1 = pipeline_builder.add_stage('Trash')
    trash2 = pipeline_builder.add_stage('Trash')
    trash3 = pipeline_builder.add_stage('Trash')
    publish_response = None
    try:
        # Intentionally avoid specifying an index value to make sure the default case works as expected
        dev_raw_data_source.connect_outputs([deduplicator])
        deduplicator.connect_outputs(stages=[trash1, trash2], output_lane_index=0)
        deduplicator.connect_outputs(stages=[trash3], output_lane_index=1)
        assert dev_raw_data_source.output_lanes == deduplicator.input_lanes
        assert trash1.input_lanes == trash2.input_lanes
        assert not (trash1.input_lanes == trash3.input_lanes or trash2.input_lanes == trash3.input_lanes)
        assert (
            trash1.input_lanes[0] in deduplicator.output_lanes
            and trash2.input_lanes[0] in deduplicator.output_lanes
            and trash3.input_lanes[0] in deduplicator.output_lanes
        )
        pipeline = pipeline_builder.build(title='connect-outputs-multi-lane-test{}'.format(get_random_string()))
        publish_response = sch.publish_pipeline(pipeline)

        # Retrieve the pipeline again and validate the same changes are seen after pushing to SCH
        pipeline = sch.pipelines.get(pipeline_id=pipeline.pipeline_id)
        dev_raw_data_source = pipeline.stages.get(instance_name='DevRawDataSource_01')
        deduplicator = pipeline.stages.get(instance_name='RecordDeduplicator_01')
        trash1 = pipeline.stages.get(instance_name='Trash_01')
        trash2 = pipeline.stages.get(instance_name='Trash_02')
        trash3 = pipeline.stages.get(instance_name='Trash_03')
        assert dev_raw_data_source.output_lanes == deduplicator.input_lanes
        assert trash1.input_lanes == trash2.input_lanes
        assert not (trash1.input_lanes == trash3.input_lanes or trash2.input_lanes == trash3.input_lanes)
        assert (
            trash1.input_lanes[0] in deduplicator.output_lanes
            and trash2.input_lanes[0] in deduplicator.output_lanes
            and trash3.input_lanes[0] in deduplicator.output_lanes
        )
    finally:
        if publish_response:
            sch.delete_pipeline(pipeline)


def test_connect_inputs_to_multi_lane_fixed_stage(sch, sch_authoring_sdc_id):
    # Building a new pipeline since we need to test the construction of one as part of this test case
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    dev_raw_data_source = pipeline_builder.add_stage('Dev Raw Data Source').set_attributes(
        data_format='JSON', raw_data='12345', stop_after_first_batch=True
    )
    deduplicator = pipeline_builder.add_stage('Record Deduplicator')
    field_flattener = pipeline_builder.add_stage('Field Flattener')
    trash1 = pipeline_builder.add_stage('Trash')
    trash2 = pipeline_builder.add_stage('Trash')
    trash3 = pipeline_builder.add_stage('Trash')
    publish_response = None
    try:
        # Intentionally avoid specifying an index value to make sure the default case works as expected
        deduplicator.connect_inputs([dev_raw_data_source])
        field_flattener.connect_inputs([dev_raw_data_source], target_stage_output_lane_index=0)
        trash1.connect_inputs(stages=[deduplicator, field_flattener], target_stage_output_lane_index=0)
        trash2.connect_inputs(stages=[deduplicator])
        trash3.connect_inputs(stages=[deduplicator], target_stage_output_lane_index=1)
        assert (
            dev_raw_data_source.output_lanes == deduplicator.input_lanes
            and dev_raw_data_source.output_lanes == field_flattener.input_lanes
        )
        assert trash2.input_lanes[0] in trash1.input_lanes
        assert (
            field_flattener.output_lanes[0] in trash1.input_lanes and deduplicator.output_lanes[0] in trash1.input_lanes
        )
        assert not (trash3.input_lanes[0] in trash1.input_lanes or trash2.input_lanes == trash3.input_lanes)
        assert trash2.input_lanes[0] in deduplicator.output_lanes and trash3.input_lanes[0] in deduplicator.output_lanes
        assert field_flattener.input_lanes == dev_raw_data_source.output_lanes
        pipeline = pipeline_builder.build(title='connect-inputs-multi-lane-test{}'.format(get_random_string()))
        publish_response = sch.publish_pipeline(pipeline)

        # Retrieve the pipeline again and validate the same changes are seen after pushing to SCH
        pipeline = sch.pipelines.get(pipeline_id=pipeline.pipeline_id)
        dev_raw_data_source = pipeline.stages.get(instance_name='DevRawDataSource_01')
        deduplicator = pipeline.stages.get(instance_name='RecordDeduplicator_01')
        field_flattener = pipeline.stages.get(instance_name='FieldFlattener_01')
        trash1 = pipeline.stages.get(instance_name='Trash_01')
        trash2 = pipeline.stages.get(instance_name='Trash_02')
        trash3 = pipeline.stages.get(instance_name='Trash_03')
        assert (
            dev_raw_data_source.output_lanes == deduplicator.input_lanes
            and dev_raw_data_source.output_lanes == field_flattener.input_lanes
        )
        assert trash2.input_lanes[0] in trash1.input_lanes
        assert (
            field_flattener.output_lanes[0] in trash1.input_lanes and deduplicator.output_lanes[0] in trash1.input_lanes
        )
        assert not (trash3.input_lanes[0] in trash1.input_lanes or trash2.input_lanes == trash3.input_lanes)
        assert trash2.input_lanes[0] in deduplicator.output_lanes and trash3.input_lanes[0] in deduplicator.output_lanes
        assert field_flattener.input_lanes == dev_raw_data_source.output_lanes
    finally:
        if publish_response:
            sch.delete_pipeline(pipeline)


def test_connect_via_legacy_rshift_operator(sch, sch_authoring_sdc_id):
    # Building a new pipeline since we need to test the construction of one as part of this test case
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    dev_raw_data_source = pipeline_builder.add_stage('Dev Raw Data Source').set_attributes(
        data_format='JSON', raw_data='12345', stop_after_first_batch=True
    )
    deduplicator = pipeline_builder.add_stage('Record Deduplicator')
    trash1 = pipeline_builder.add_stage('Trash')
    trash2 = pipeline_builder.add_stage('Trash')
    trash3 = pipeline_builder.add_stage('Trash')
    publish_response = None
    try:
        dev_raw_data_source >> deduplicator >> [trash1, trash2]
        deduplicator >> trash3
        assert dev_raw_data_source.output_lanes == deduplicator.input_lanes
        assert trash1.input_lanes[0] == trash2.input_lanes[0] == deduplicator.output_lanes[0]
        assert trash3.input_lanes[0] == deduplicator.output_lanes[1]
        pipeline = pipeline_builder.build(title='connect-legacy-multi-lane-test{}'.format(get_random_string()))
        publish_response = sch.publish_pipeline(pipeline)

        # Retrieve the pipeline again and validate the same changes are seen after pushing to SCH
        pipeline = sch.pipelines.get(pipeline_id=pipeline.pipeline_id)
        dev_raw_data_source = pipeline.stages.get(instance_name='DevRawDataSource_01')
        deduplicator = pipeline.stages.get(instance_name='RecordDeduplicator_01')
        trash1 = pipeline.stages.get(instance_name='Trash_01')
        trash2 = pipeline.stages.get(instance_name='Trash_02')
        trash3 = pipeline.stages.get(instance_name='Trash_03')
        assert dev_raw_data_source.output_lanes == deduplicator.input_lanes
        assert trash1.input_lanes[0] == trash2.input_lanes[0] == deduplicator.output_lanes[0]
        assert trash3.input_lanes[0] == deduplicator.output_lanes[1]
    finally:
        if publish_response:
            sch.delete_pipeline(pipeline)


@pytest.mark.selector
def test_stream_selector_with_different_initial_value_format(pipeline_builder):
    selector = pipeline_builder.add_stage('Stream Selector')

    assert len(selector.output_lanes) == 1
    assert selector.condition[0].get('predicate', None) == 'default'
    assert all([s.get('outputLane', False) for s in selector.condition])

    predicate = '>=0'
    selector.condition = [predicate]
    assert len(selector.output_lanes) == 2
    assert selector.condition[0].get('predicate', None) == predicate
    assert selector.condition[1].get('predicate', None) == 'default'
    assert all([s.get('outputLane', False) for s in selector.condition])
    assert selector.output_streams == len(selector.output_lanes)

    selector.condition = [{'predicate': predicate}]
    assert len(selector.output_lanes) == 2
    assert selector.condition[0].get('predicate', None) == predicate
    assert selector.condition[1].get('predicate', None) == 'default'
    assert all([s.get('outputLane', False) for s in selector.condition])
    assert selector.output_streams == len(selector.output_lanes)

    selector.condition = [{'predicate': predicate, 'outputLane': 'this'}, {'predicate': 'default'}]
    assert len(selector.output_lanes) == 2
    assert selector.condition[0].get('predicate', None) == predicate
    assert selector.condition[1].get('predicate', None) == 'default'
    assert all([s.get('outputLane', False) for s in selector.condition])
    assert selector.output_streams == len(selector.output_lanes)

    selector.condition = [{'predicate': predicate}, 'default']
    assert len(selector.output_lanes) == 2
    assert selector.condition[0].get('predicate', None) == predicate
    assert selector.condition[1].get('predicate', None) == 'default'
    assert all([s.get('outputLane', False) for s in selector.condition])
    assert selector.output_streams == len(selector.output_lanes)


@pytest.mark.selector
def test_invalid_predicate_format(pipeline_builder):
    selector = pipeline_builder.add_stage('Stream Selector')

    with pytest.raises(ValueError):
        selector.condition = ['default', '>0']

    with pytest.raises(ValueError):
        selector.condition = [{'predicate': 'default'}, {'predicate': '>0'}]

    with pytest.raises(ValueError):
        selector.add_predicates('single predicate')

    with pytest.raises(ValueError):
        selector.add_predicates([{'invalidkey': 'invalid'}])


@pytest.mark.selector
def test_output_lanes_are_different_for_stream_selector(pipeline_builder):
    num_conditions = 10
    selector_stage = pipeline_builder.add_stage('Stream Selector')
    selector_stage.condition = ['>0' for i in range(num_conditions)]

    for i in range(num_conditions):
        for j in range(num_conditions):
            if i != j:
                assert selector_stage.output_lanes[i] != selector_stage.output_lanes[j]


@pytest.mark.selector
def test_stream_selector_connections(pipeline_builder):
    selector_stage = pipeline_builder.add_stage('Stream Selector')
    trash = pipeline_builder.add_stage('Trash')

    selector_stage >> trash

    assert selector_stage.output_lanes[0] == selector_stage.condition[0]['outputLane']
    assert selector_stage.output_lanes[0] == trash.input_lanes[0]

    selector_stage.condition = ['>=0']
    # Assert that the trash was disconnected.
    assert len(trash.input_lanes) == 0
    assert len(selector_stage.output_lanes) == 2

    trash2 = pipeline_builder.add_stage('Trash')
    selector_stage >> trash
    selector_stage >> trash2

    # Assert the proper order for connections
    assert selector_stage.output_lanes[0] == selector_stage.condition[0]['outputLane']
    assert selector_stage.output_lanes[0] == trash.input_lanes[0]
    assert selector_stage.output_lanes[1] == selector_stage.condition[1]['outputLane']
    assert selector_stage.output_lanes[1] == trash2.input_lanes[0]

    # Assert both trashes get disconnected
    selector_stage.condition = []
    assert len(trash.input_lanes) == 0
    assert len(trash2.input_lanes) == 0


@pytest.mark.selector
def test_publish_pipeline_with_stream_selector(engine_type, pipeline_builder, sch):
    if engine_type == 'snowflake':
        src_stage = pipeline_builder.add_stage('Snowflake Table', type='origin')
    else:
        src_stage = pipeline_builder.add_stage('Dev Raw Data Source')

    stream_selector = pipeline_builder.add_stage('Stream Selector')

    trash1 = pipeline_builder.add_stage('Trash')
    trash2 = pipeline_builder.add_stage('Trash')

    stage_names = [stage.instance_name for stage in [src_stage, stream_selector, trash1, trash2]]

    stream_selector.condition = ['>=0', 'default']

    src_stage >> stream_selector
    stream_selector >> trash1
    stream_selector >> trash2

    pipeline_name = 'simple_pipeline_with_stream_selector_{}'.format(get_random_string())
    pipeline = pipeline_builder.build(pipeline_name)
    try:
        sch.publish_pipeline(pipeline)
        published_pipeline = sch.pipelines.get(name=pipeline_name)
        src_stage, stream_selector, trash1, trash2 = [
            published_pipeline.stages.get(instance_name=name) for name in stage_names
        ]
        assert published_pipeline is not None
        assert stream_selector.output_lanes[0] == trash1.input_lanes[0]
        assert stream_selector.output_lanes[1] == trash2.input_lanes[0]
    finally:
        sch.delete_pipeline(pipeline)


@pytest.mark.selector
def test_edit_pipeline_with_stream_selector(engine_type, pipeline_builder, sch):
    if engine_type == 'snowflake':
        src_stage = pipeline_builder.add_stage('Snowflake Table', type='origin')
    else:
        src_stage = pipeline_builder.add_stage('Dev Raw Data Source')

    stream_selector = pipeline_builder.add_stage('Stream Selector')

    trash1 = pipeline_builder.add_stage('Trash')
    trash2 = pipeline_builder.add_stage('Trash')

    stage_names = [stage.instance_name for stage in [src_stage, stream_selector, trash1, trash2]]

    stream_selector.condition = ['>=0', 'default']

    src_stage >> stream_selector
    stream_selector >> trash1
    stream_selector >> trash2

    pipeline_name = 'simple_pipeline_with_stream_selector_{}'.format(get_random_string())
    pipeline = pipeline_builder.build(pipeline_name)

    try:
        sch.publish_pipeline(pipeline)
        pipeline = sch.pipelines.get(name=pipeline_name)
        src_stage, stream_selector, trash1, trash2 = [pipeline.stages.get(instance_name=name) for name in stage_names]
        stream_selector.condition = ['=0']
        assert stream_selector.condition[0]['predicate'] == '=0'
        assert stream_selector.condition[1]['predicate'] == 'default'
        assert not trash1.input_lanes
        assert not trash2.input_lanes
    finally:
        sch.delete_pipeline(pipeline)


@pytest.mark.selector
def test_stream_selector_has_output_lanes_with_different_id(pipeline_builder):
    stream_selector = pipeline_builder.add_stage('Stream Selector')

    stream_selector.predicates = ['0']
    lanes = stream_selector.output_lanes
    assert len(lanes) == 2
    # Check that no lane repeats
    assert len(lanes) == len(set(lanes))
    assert all(isinstance(predicate, dict) for predicate in stream_selector.predicates)

    stream_selector.add_predicates(['1'])
    stream_selector.add_predicates(['2'])

    lanes = stream_selector.output_lanes
    assert len(lanes) == 4
    # Check that no lane repeats
    assert len(lanes) == len(set(lanes))


@pytest.mark.selector
def test_remove_predicate_from_stream_selector(engine_type, pipeline_builder):
    if engine_type == 'snowflake':
        src_stage = pipeline_builder.add_stage('Snowflake Table', type='origin')
    else:
        src_stage = pipeline_builder.add_stage('Dev Raw Data Source')

    stream_selector = pipeline_builder.add_stage('Stream Selector')

    trash1 = pipeline_builder.add_stage('Trash')
    trash2 = pipeline_builder.add_stage('Trash')

    stream_selector.condition = ['>=0', 'default']

    src_stage >> stream_selector
    stream_selector >> trash1
    stream_selector >> trash2

    assert stream_selector.output_lanes[0] == trash1.input_lanes[0]
    assert len(stream_selector.output_lanes) == stream_selector.output_streams

    stream_selector.remove_predicate(stream_selector.condition[0])

    assert len(trash1.input_lanes) == 0
    assert len(stream_selector.output_lanes) == 1
    assert stream_selector.output_lanes[0] == trash2.input_lanes[0]
    assert len(stream_selector.output_lanes) == stream_selector.output_streams


@pytest.mark.selector
def test_add_predicates_to_stream_selector(engine_type, pipeline_builder):
    stream_selector = pipeline_builder.add_stage('Stream Selector')

    stream_selector.condition = ['default']
    assert len(stream_selector.predicates) == 1
    assert len(stream_selector.output_lanes) == len(stream_selector.predicates)

    stream_selector.add_predicates(['>=0'])

    assert len(stream_selector.predicates) == 2
    assert stream_selector.predicates[0]['predicate'] == '>=0'
    assert stream_selector.predicates[1]['predicate'] == 'default'
    assert len(stream_selector.output_lanes) == len(stream_selector.predicates)
    assert len(stream_selector.output_lanes) == stream_selector.output_streams


def test_add_and_remove_fragment_from_pipeline(sch, sch_authoring_sdc_id, sample_fragment):
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    trash = pipeline_builder.add_stage('Trash')
    pipeline = pipeline_builder.build('sample_pipeline_{}'.format(get_random_string()))
    number_of_stages = 2
    number_of_fragments = 1

    fragment_stage = pipeline.add_fragment(sample_fragment)
    fragment_stage >> trash

    # check that fragment has been added to stages and the internal fragments list
    assert len(pipeline.stages) == number_of_stages
    assert len(pipeline._pipeline_definition['fragments']) == number_of_fragments

    # test that the fragment metadata has been added to pipeline._builder
    assert len(pipeline._builder._fragment_commit_ids) == number_of_fragments
    assert pipeline._builder._fragment_commit_ids[0] == sample_fragment.commit_id
    assert pipeline._builder._fragment_instance_count[fragment_stage._data['uiInfo']['fragmentName']] == 1

    sch.publish_pipeline(pipeline)

    try:
        pipeline = sch.pipelines.get(name=pipeline.name)
        fragment_stage = pipeline.stages.get(instance_name=fragment_stage.instance_name)

        # after pulling from platform, check that fragment has been added to stages and the internal fragments list
        assert len(pipeline.stages) == number_of_stages
        assert len(pipeline._pipeline_definition['fragments']) == number_of_fragments

        # test that the remove functionality works
        pipeline.remove_stages(fragment_stage)

        # check that fragment has been removed from stages and the internal fragments list
        assert len(pipeline.stages) == number_of_stages - number_of_fragments
        assert len(pipeline._pipeline_definition['fragments']) == number_of_fragments - 1
        assert pipeline.stages[0].instance_name == trash.instance_name

        # test that the fragment metadata has been removed from pipeline._builder
        pipeline._get_builder()
        assert len(pipeline._builder._fragment_commit_ids) == number_of_fragments - 1
        assert pipeline._builder._fragment_instance_count[fragment_stage._data['uiInfo']['fragmentName']] == 0

    finally:
        sch.delete_pipeline(pipeline)


def test_copy_inputs_and_copy_outputs(sch, sample_pipeline):
    dev, trash = sample_pipeline.stages[0], sample_pipeline.stages[-1]
    field_renamer = sample_pipeline.add_stage('Field Renamer')

    field_renamer.copy_inputs(trash)
    assert field_renamer.input_lanes[0] == trash.input_lanes[0]

    field_renamer.copy_outputs(dev)
    assert field_renamer._data['outputLanes'][0] in trash._data['inputLanes']
    assert dev._data['outputLanes'][0] in trash._data['inputLanes']

    sch.publish_pipeline(sample_pipeline)
    sample_pipeline = sch.pipelines.get(pipeline_id=sample_pipeline.pipeline_id)
    dev = sample_pipeline.stages.get(instance_name='DevDataGenerator_01')
    trash = sample_pipeline.stages.get(instance_name='Trash_01')
    field_renamer = sample_pipeline.stages.get(instance_name='FieldRenamer_01')

    assert field_renamer.input_lanes[0] == trash.input_lanes[0]
    assert field_renamer._data['outputLanes'][0] in trash._data['inputLanes']
    assert dev._data['outputLanes'][0] in trash._data['inputLanes']



# Copyright 2023 StreamSets Inc.


def test_ec2_deployment(test_aws_environment, sch):
    deployment_builder = sch.get_deployment_builder(deployment_type='EC2')
    deployment = deployment_builder.build(
        deployment_name='Sample Test EC2 Deployment',
        environment=test_aws_environment,
        engine_type='DC',
        engine_version='4.1.0',
        deployment_tags=['ec2-deployment-tag'],
    )
    deployment.desired_instances = 1
    deployment.ec2_instance_type = 'm4.large'
    deployment.aws_tags = {}
    sch.add_deployment(deployment)

    assert deployment.is_complete()
    sch.delete_deployment(deployment)



# Copyright 2022 StreamSets Inc.


def test_pipeline_parameters(sch, sch_authoring_sdc_id):
    try:
        pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
        dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
        trash = pipeline_builder.add_stage('Trash')
        dev_data_generator >> trash
        pipeline = pipeline_builder.build('pipeline_parameters_test_pipeline')

        # Set
        params = {'abc': '456', 'def': '123'}
        pipeline.parameters = params
        sch.publish_pipeline(pipeline)
        pipeline_new = sch.pipelines.get(pipeline_id=pipeline.pipeline_id)
        assert dict(pipeline_new.parameters) == params

        # Update
        params_to_be_updated = {'abc': '123', 'ghi': '123'}
        pipeline_new.parameters.update(params_to_be_updated)
        sch.publish_pipeline(pipeline_new)
        pipeline_new = sch.pipelines.get(pipeline_id=pipeline_new.pipeline_id)
        params.update(params_to_be_updated)
        assert dict(pipeline_new.parameters) == params
    finally:
        if pipeline:
            sch.delete_pipeline(pipeline)



# Copyright 2022 StreamSets Inc.

# fmt: off
import pytest

from streamsets.sdk.sch_models import ConnectionVerificationResult

# fmt: on


def test_verify_invalid_connection(sch, invalid_jdbc_test_connection):
    verification = sch.verify_connection(invalid_jdbc_test_connection)
    assert verification.status == 'INVALID'


def test_connection_restriction_for_s3_stage(
    sch, sch_authoring_sdc_id, invalid_jdbc_test_connection, invalid_aws_test_connection
):
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    origin = pipeline_builder.add_stage('Dev Data Generator')
    destination = pipeline_builder.add_stage('Amazon S3', type='destination')
    origin >> destination
    with pytest.raises(ValueError):
        destination.use_connection(invalid_jdbc_test_connection)
    assert destination.connection == 'MANUAL'
    destination.use_connection(invalid_aws_test_connection)
    assert destination.connection == invalid_aws_test_connection.id


def test_connection_restriction_for_jdbc_stage(
    sch, sch_authoring_sdc_id, invalid_jdbc_test_connection, invalid_aws_test_connection
):
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    origin = pipeline_builder.add_stage('Dev Data Generator')
    destination = pipeline_builder.add_stage('JDBC Producer')
    origin >> destination
    with pytest.raises(ValueError):
        destination.use_connection(invalid_aws_test_connection)
    assert destination.connection == 'MANUAL'
    destination.use_connection(invalid_jdbc_test_connection)
    assert destination.connection == invalid_jdbc_test_connection.id


def test_snowflake_connection(sch, sch_authoring_sdc_id, invalid_snowflake_test_connection):
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    snowflake_destination = pipeline_builder.add_stage('Snowflake', type='destination')

    assert snowflake_destination.connection == 'MANUAL'
    snowflake_destination.use_connection(invalid_snowflake_test_connection)
    assert snowflake_destination.connection == invalid_snowflake_test_connection.id


def test_engineless_connection_restriction_for_s3_stage(
    sch, sch_authoring_sdc_id, invalid_jdbc_engineless_test_connection, invalid_aws_engineless_test_connection
):
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    origin = pipeline_builder.add_stage('Dev Data Generator')
    destination = pipeline_builder.add_stage('Amazon S3', type='destination')
    origin >> destination
    with pytest.raises(ValueError):
        destination.use_connection(invalid_jdbc_engineless_test_connection)
    assert destination.connection == 'MANUAL'
    destination.use_connection(invalid_aws_engineless_test_connection)
    assert destination.connection == invalid_aws_engineless_test_connection.id


def test_engineless_connection_restriction_for_jdbc_stage(
    sch, sch_authoring_sdc_id, invalid_jdbc_engineless_test_connection, invalid_aws_engineless_test_connection
):
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    origin = pipeline_builder.add_stage('Dev Data Generator')
    destination = pipeline_builder.add_stage('JDBC Producer')
    origin >> destination
    with pytest.raises(ValueError):
        destination.use_connection(invalid_aws_engineless_test_connection)
    assert destination.connection == 'MANUAL'
    destination.use_connection(invalid_jdbc_engineless_test_connection)
    assert destination.connection == invalid_jdbc_engineless_test_connection.id


def test_engineless_snowflake_connection(sch, sch_authoring_sdc_id, invalid_snowflake_engineless_test_connection):
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    snowflake_destination = pipeline_builder.add_stage('Snowflake', type='destination')

    assert snowflake_destination.connection == 'MANUAL'
    snowflake_destination.use_connection(invalid_snowflake_engineless_test_connection)
    assert snowflake_destination.connection == invalid_snowflake_engineless_test_connection.id


@pytest.mark.parametrize('unsupported_test_connection', ['invalid_jdbc_test_connection', 'invalid_aws_test_connection'])
def test_connection_restriction_for_snowflake_stage(sch, sch_authoring_sdc_id, unsupported_test_connection, request):
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    snowflake_destination = pipeline_builder.add_stage('Snowflake', type='destination')
    unsupported_test_connection = request.getfixturevalue(unsupported_test_connection)

    with pytest.raises(ValueError):
        snowflake_destination.use_connection(unsupported_test_connection)


@pytest.mark.parametrize(
    "library_to_check_against, use_sdc_snowflake_library",
    [(None, False), (None, True), ('jdbc', False)],  # no params  # snowflake library  # check against jdbc library
)
def test_verify_connection(
    sch, invalid_jdbc_test_connection, library_to_check_against, sdc_snowflake_library, use_sdc_snowflake_library
):
    if use_sdc_snowflake_library:
        # since we pass the version of the snowflake library we should remove the version from it
        library = sdc_snowflake_library.split(':')[0]
    else:
        library = library_to_check_against

    connection_verification_result = sch.verify_connection(
        connection=invalid_jdbc_test_connection,
        library=library,
    )

    assert isinstance(connection_verification_result, ConnectionVerificationResult)



# Copyright 2019 StreamSets Inc.

"""Common exceptions."""

# fmt: off
from json.decoder import JSONDecodeError

from requests.exceptions import HTTPError

# fmt: on


class InternalServerError(HTTPError):
    """Internal server error."""

    def __init__(self, response):
        self.status_code = response.status_code
        try:
            self.response = response.json()
            self.text = self.response.get('RemoteException', {}).get('localizedMessage')

            if self.text is None:
                # if text is None, and we did not get a JSONDecodeError,
                # there is an error message in the JSON that is not formatted as expected
                self.text = str(self.response)

        except JSONDecodeError:
            self.text = response.text

        # Propagate message from the REST interface as message of the exception
        super().__init__(self.text)


class BadRequestError(HTTPError):
    """Bad request error (HTTP 400)."""

    def __init__(self, response):
        super().__init__(response.text)


class LegacyDeploymentInactiveError(Exception):
    """Legacy deployment status changed into INACTIVE_ERROR."""

    def __init__(self, message):
        self.message = message


class ValidationError(Exception):
    """Validation issues."""

    def __init__(self, issues):
        self.issues = issues


class JobInactiveError(Exception):
    """Job status changed into INACTIVE_ERROR."""

    def __init__(self, message):
        self.message = message


class JobRunnerError(Exception):
    """JobRunner errors."""

    def __init__(self, code, message):
        super().__init__('{}: {}'.format(code, message))
        self.code = code
        self.message = message


class UnsupportedMethodError(Exception):
    """An unsupported method was called."""

    def __init__(self, message):
        self.message = message


class TopologyIssuesError(Exception):
    """Topology has some issues."""

    def __init__(self, issues):
        self.message = issues


class InvalidCredentialsError(Exception):
    """Invalid credentials error."""

    def __init__(self, message):
        self.message = message


class ConnectionError(Exception):
    """Connection Catalog errors."""

    def __init__(self, code, message):
        super().__init__('{}: {}'.format(code, message))
        self.message = message
        self.code = code


class MultipleIssuesError(Exception):
    """Multiple errors were returned."""

    def __init__(self, errors):
        self.errors = errors
        self.message = 'Multiple issues encountered'
        super().__init__('{}: {}'.format(self.message, self.errors))

    def __getitem__(self, item):
        return self.errors[item]

    def __repr__(self):
        return str('{}: {}'.format(self.message, self.errors))


class UnprocessableEntityError(HTTPError):
    """Unprocessable Entity Error (HTTP 422)."""

    def __init__(self, response):
        self.response = response.json()
        # The error response from ASTER's API puts the error in a list, so we index into the list to get the message
        self.message = self.response['errors'][0]['message']
        self.code = self.response['status']

        # Propagate the message from the API interface
        super().__init__('{}: {}'.format(self.code, self.message))


class InvalidVersionError(ValueError):
    """The Version number could not be parsed."""

    def __init__(self, input):
        self.input = input
        super().__init__("{!r} is not a valid version.".format(self.input))


class StatusError(Exception):
    """Parent class for pipeline status errors."""

    def __init__(self, response):
        self.response = response
        self.message = self.response.get('message')
        if self.message:
            super().__init__(self.message)
        else:
            self.message = 'Unknown status message'
            super().__init__(response)


class StartError(StatusError):
    """Pipeline start error."""

    pass


class StartingError(StatusError):
    """Pipeline starting error."""

    pass


class RunError(StatusError):
    """Pipeline running error."""

    pass


class RunningError(StatusError):
    """Pipeline run error."""

    pass


class ConnectError(StatusError):
    """Pipeline connect error."""

    pass


class InvalidError(StatusError):
    """Invalid stage config error."""

    pass


class ServiceDefinitionNotFound(Exception):
    """ServiceDefinition errors."""

    def __init__(self, message):
        super().__init__('{}'.format(message))
        self.message = message


class EnginelessError(Exception):
    """Publishing an engineless designed pipeline error"""

    def __init__(self, message):
        self.message = message



# Copyright 2021 StreamSets Inc.

"""Abstractions for interacting with StreamSets Transformer."""

# fmt: off
import copy
import io
import json
import logging
import time
import zipfile
from datetime import datetime
from functools import wraps
from urllib.parse import urlparse

import requests

from . import st_api, st_models
from .constants import ENGINE_AUTHENTICATION_METHOD_FORM
from .exceptions import ValidationError
from .utils import SeekableList

# fmt: on

logger = logging.getLogger(__name__)

# The `#:` constructs at the end of assignments are part of Sphinx's autodoc functionality.
DEFAULT_START_STATUSES_TO_WAIT_FOR = ['RUNNING', 'FINISHED']  #:
DEFAULT_START_TIMEOUT = 300  #:
DEFAULT_STOP_TIMEOUT = 300  #:
DEFAULT_WAIT_FOR_METRIC_TIMEOUT = 30  #:

# the server side for a validate (i.e. when the runner will give up on waiting for the validate to finish)
DEFAULT_VALIDATE_SERVER_TIMEOUT_SEC = 500
# client side validate timeout
DEFAULT_VALIDATE_CLIENT_TIMEOUT_SEC = 500
DEFAULT_VALIDATE_TIME_BETWEEN_CHECKS = 2

# the server side for a preview (i.e. when the runner will give up on waiting for the preview to finish)
DEFAULT_PREVIEW_SERVER_TIMEOUT_MS = 120000
# for now, make the client side default timeout the same value as the server side
DEFAULT_PREVIEW_CLIENT_TIMEOUT_SEC = DEFAULT_PREVIEW_SERVER_TIMEOUT_MS / 1000
DEFAULT_PREVIEW_TIME_BETWEEN_CHECKS = 2


class Transformer:
    """Class to interact with StreamSets Transformer.

    Args:
        server_url (:obj:`str`): URL of an existing ST deployment with which to interact.
        aster_authentication_token (:obj:`str`, optional): Aster authentication token. Default: ``None``
        aster_server_url (:obj:`str`, optional): Aster server base URL. Default: ``None``
        authentication_method (:obj:`str`, optional): StreamSets Transformer authentication method.
            Default: :py:const:`streamsets.sdk.constants.ENGINE_AUTHENTICATION_METHOD_FORM`.
        control_hub (:py:class:`streamsets.sdk.ControlHub`, optional): ControlHub instance to use
            for ControlHub-registered Transformers. Default: ``None``.
        dump_log_on_error (:obj:`bool`): Whether to output Transformer logs when exceptions
            are raised by certain methods. Default: ``False``
    """

    VERIFY_SSL_CERTIFICATES = True

    def __init__(
        self,
        server_url,
        aster_authentication_token=None,
        aster_server_url=None,
        authentication_method=ENGINE_AUTHENTICATION_METHOD_FORM,
        control_hub=None,
        dump_log_on_error=False,
        **kwargs
    ):
        self.server_url = server_url
        self.authentication_method = authentication_method
        self.aster_authentication_token = aster_authentication_token
        self.aster_server_url = aster_server_url
        self.control_hub = control_hub
        self.dump_log_on_error = dump_log_on_error

        if self.server_url:
            sch_headers = (
                {
                    'X-SS-App-Component-Id': self.control_hub.api_client.session.headers['X-SS-App-Component-Id'],
                    'X-SS-App-Auth-Token': self.control_hub.api_client.session.headers['X-SS-App-Auth-Token'],
                }
                if self.control_hub
                else {}
            )
            session_attributes = {'verify': self.VERIFY_SSL_CERTIFICATES}
            if self.control_hub.use_websocket_tunneling:
                current_server_url = '{}/tunneling/rest/{}'.format(control_hub.server_url, kwargs.get('transformer_id'))
            else:
                current_server_url = self.server_url
            self.api_client = st_api.ApiClient(
                server_url=current_server_url,
                authentication_method=self.authentication_method,
                aster_authentication_token=aster_authentication_token,
                aster_server_url=aster_server_url,
                headers=sch_headers,
                dump_log_on_error=self.dump_log_on_error,
                session_attributes=session_attributes,
                **kwargs
            )

            # Keep track of the server host so that tests that may need it (e.g. to set configurations) can use it.
            self.server_host = urlparse(self.server_url).netloc.split(':')[0]
        else:
            self.server_host = None
            self.api_client = None

        # ST definitions should be an attribute of this class, but we use a property for
        # access to handle necessary setup and synchronization tasks, so indicate internal use for
        # the underlying attribute with a leading underscore.
        self._definitions = None

        # Instances of :py:class:`streamsets.sdk.st_models.PipelineBuilder` require a blank pipeline (i.e.
        # an empty pipeline JSON) from our particular Transformer version. We keep track of this with an
        # instance attribute to avoid repetitive API calls.
        self._pipeline = None

    def _dump_st_log_on_error(*dec_args, **dec_kwargs):
        """A Python decorator to log ST when errors happen.

        Args:
            *dec_args: Optional positional arguments to be passed.
            **dec_kwargs: Optional keyword arguments to be passed, such as ``all`. ``all`` will
                include complete ST logs.
        """

        def outer_func(func):
            @wraps(func)
            def wrapped(self, *args, **kwargs):
                log_time_now = datetime.now().strftime('%Y-%m-%d %H:%M:%S,%f')
                try:
                    return func(self, *args, **kwargs)
                except Exception:
                    if self.dump_log_on_error:
                        st_log = self.get_logs() if dec_kwargs.get('all') else self.get_logs().after_time(log_time_now)
                        if st_log:
                            logger.error('Error during `%s` call. ST log follows ...', func.__name__)
                            print('------------------------- ST log - Begins -----------------------')
                            print(st_log)
                            print('------------------------- ST log - Ends -------------------------')
                    raise

            return wrapped

        if len(dec_args) == 1 and not dec_kwargs and callable(dec_args[0]):  # called without args
            return outer_func(dec_args[0])
        else:
            return outer_func

    @property
    def version(self):
        """Return the version of the Transformer.

        Returns:
           :obj:`str`: The version string.
        """
        return self.api_client.get_st_info()['version']

    @_dump_st_log_on_error
    def add_pipeline(self, *pipelines):
        """Add one or more pipelines to the Transformer instance.

        Args:
            *pipelines: One or more instances of :py:class:`streamsets.sdk.st_models.Pipeline`.
        """
        for pipeline in set(pipelines):
            # Only do the REST call to add the pipeline if an API client is available.
            if self.api_client:
                logger.info('Importing pipeline %s...', pipeline.id)
                response = self.api_client.import_pipeline(
                    pipeline_id=pipeline.id, pipeline_json=pipeline._data, auto_generate_pipeline_id=True
                )
                pipeline._data = response
                status_command = self.api_client.get_pipeline_status(pipeline_id=pipeline.id)
                status_command.wait_for_status(status='EDITED')

    @_dump_st_log_on_error
    def remove_pipeline(self, *pipelines):
        """Remove one or more pipelines from the Transformer instance.

        Args:
            *pipelines: One or more instances of :py:class:`streamsets.sdk.st_models.Pipeline`.
        """
        for pipeline in set(pipelines):
            if self.api_client:
                logger.info('Deleting pipeline %s...', pipeline.id)
                self.api_client.delete_pipeline(pipeline_id=pipeline.id)

    def reset_origin(self, pipeline):
        """Reset origin offset.

        Args:
            pipeline (:py:class:`streamsets.sdk.st_models.Pipeline`): Pipeline object.

        Returns:
            An instance of :py:class:`streamsets.sdk.st_api.Command`.
        """
        return self.api_client.reset_origin_offset(pipeline.id)

    @property
    def pipelines(self):
        """Get all pipelines in the pipeline store.

        Returns:
            A :py:obj:`streamsets.sdk.utils.SeekableList` of
                :py:class:`streamsets.sdk.st_models.Pipeline` instances.
        """
        pipeline_ids = [
            pipeline_configuration_info['pipelineId']
            for pipeline_configuration_info in self.api_client.get_all_pipeline_configuration_info().response.json()
        ]
        pipelines_archive = zipfile.ZipFile(
            file=io.BytesIO(self.api_client.export_pipelines(body=pipeline_ids).response.content), mode='r'
        )
        pipelines = [
            json.loads(pipelines_archive.read(pipeline_filename).decode())
            for pipeline_filename in pipelines_archive.namelist()
        ]
        return SeekableList(
            self.get_pipeline_builder()
            .import_pipeline(pipeline, regenerate_id=False)
            .build(title=pipeline['pipelineConfig']['title'])
            for pipeline in pipelines
        )

    def get_pipeline_builder(self, **kwargs):
        """Get a pipeline builder instance with which a pipeline can be created.

        Returns:
            An instance of :py:class:`streamsets.sdk.st_models.PipelineBuilder`.
        """
        if not self.api_client:
            raise Exception('ST must be started to get a PipelineBuilder instance.')

        # A :py:class:`streamsets.sdk.st_models.PipelineBuilder` instance takes an empty pipeline and a
        # dictionary of definitions as arguments. To get the former, we generate a pipeline in ST, export it,
        # and then delete it. For the latter, we simply pass along `self.definitions`.
        if not self._pipeline:
            fragment = kwargs.get('fragment', False)
            if not fragment:
                create_pipeline_response = self.api_client.create_pipeline(
                    pipeline_title='Pipeline Builder', auto_generate_pipeline_id=True
                ).response.json()
            else:
                execution_mode = kwargs.get('execution_mode', 'STANDALONE')
                title = 'Pipeline_Fragment_Builder'
                description = 'Pipeline_Fragment_Builder_Description'
                draft = kwargs.get('draft', True)
                create_pipeline_response = self.api_client.create_pipeline_fragment(
                    pipeline_fragment_title=title, execution_mode=execution_mode, description=description, draft=draft
                ).response.json()
            try:
                pipeline_id = create_pipeline_response['info']['pipelineId']
            except KeyError:
                pipeline_id = create_pipeline_response['info']['name']

            self._pipeline = self.api_client.export_pipeline(pipeline_id)
            self.api_client.delete_pipeline(pipeline_id)

        return st_models.PipelineBuilder(pipeline=copy.deepcopy(self._pipeline), definitions=self.definitions)

    @property
    def definitions(self):
        """Get an ST instance's definitions.

        Will return a cached instance of the definitions if called more than once.

        Returns:
            An instance of :py:class:`json`.
        """
        if self._definitions:
            return self._definitions

        # Getting definitions from ST requires a running deployment.
        if not self.api_client:
            raise Exception('ST must be started to get definitions.')

        self._definitions = self.api_client.get_definitions()
        return self._definitions

    @_dump_st_log_on_error
    def validate_pipeline(self, pipeline, **kwargs):
        """Validate a pipeline.

        Args:
            pipeline (:py:class:`streamsets.sdk.st_models.Pipeline`): The pipeline instance.
            timeout (:obj:`int`, optional): Server side validate Timeout in seconds.
                Default: :py:const:`streamsets.sdk.st.DEFAULT_VALIDATE_SERVER_TIMEOUT_SEC`.
            timeout_sec (:obj:`int`, optional): Client side validate timeout, in seconds.
                Default: :py:const:`streamsets.sdk.st.DEFAULT_VALIDATE_CLIENT_TIMEOUT_SEC`.
            time_between_checks (:obj:`int`, optional): Time to sleep between validation checks.
                Default: :py:const:`streamsets.sdk.st.DEFAULT_VALIDATE_TIME_BETWEEN_CHECKS`.
            using_configured_cluster_manager (:obj:`bool`, optional): Validate pipeline using configured cluster
                manager. Default: ``True``.
        """
        logger.info('Validating pipeline %s ...', pipeline.id)
        server_timeout_ms = kwargs.get('timeout', DEFAULT_VALIDATE_SERVER_TIMEOUT_SEC) * 1000
        timeout_sec = kwargs.get('timeout_sec', DEFAULT_VALIDATE_CLIENT_TIMEOUT_SEC)
        time_between_checks = kwargs.get('time_between_checks', DEFAULT_VALIDATE_TIME_BETWEEN_CHECKS)
        using_configured_cluster_manager = kwargs.get('using_configured_cluster_manager', True)

        if time_between_checks > timeout_sec:
            raise ValueError('Time to sleep between validate status check cannot be greater than validate timeout')

        validate_command = self.api_client.validate_pipeline(
            pipeline_id=pipeline.id, timeout=server_timeout_ms, remote=using_configured_cluster_manager
        ).wait_for_validate(timeout_sec=timeout_sec, time_between_checks=time_between_checks)
        response = validate_command.response.json()
        if response['status'] != 'VALID':
            if 'issues' in response and response['issues']:
                raise ValidationError(response['issues'])
            elif 'message' in response:
                raise ValidationError(response['message'])
            else:
                raise ValidationError('Unknown validation failure with status as {}'.format(response['status']))

    @_dump_st_log_on_error
    def start_pipeline(self, pipeline_id, runtime_parameters=None, **kwargs):
        """Start a pipeline.

        Args:
            pipeline_id (:obj:`str`): The pipeline instance's ID.
            runtime_parameters (:obj:`dict`, optional): Collection of runtime parameters. Default: ``None``.
            wait (:obj:`bool`, optional): Wait for pipeline to start. Default: ``True``.
            wait_for_statuses (:obj:`list`, optional): Pipeline statuses to wait on.
                Default: ``['RUNNING', 'FINISHED']``.
            timeout_sec (:obj:`int`): Timeout to wait for pipeline statuses, in seconds.
                Default: :py:const:`streamsets.sdk.st.DEFAULT_START_TIMEOUT`.

        Returns:
            An instance of :py:class:`streamsets.sdk.st_api.PipelineCommand`.
        """
        issues = self.api_client.get_pipeline_configuration(pipeline_id=pipeline_id)['issues']
        if issues['issueCount']:
            raise ValidationError(issues)

        logger.info('Starting pipeline %s ...', pipeline_id)
        pipeline_command = self.api_client.start_pipeline(
            pipeline_id=pipeline_id, runtime_parameters=runtime_parameters
        )
        if kwargs.get('wait', True):
            timeout_sec = kwargs.get('timeout_sec', DEFAULT_START_TIMEOUT)
            pipeline_command.wait_for_status(
                status=kwargs.get('wait_for_statuses', DEFAULT_START_STATUSES_TO_WAIT_FOR), timeout_sec=timeout_sec
            )

        return pipeline_command

    @_dump_st_log_on_error
    def stop_pipeline(self, pipeline, **kwargs):
        """Stop a pipeline.

        Args:
            pipeline (:py:class:`streamsets.sdk.st_models.Pipeline`): The pipeline instance.
            force (:obj:`bool`, optional): Force pipeline to stop. Default: ``False``.
            wait (:obj:`bool`, optional): Wait for pipeline to stop. Default: ``True``.
            timeout_sec (:obj:`int`): Timeout to wait for pipeline stop, in seconds.
                Default: :py:const:`streamsets.sdk.st.DEFAULT_STOP_TIMEOUT`.

        Returns:
            An instance of :py:class:`streamsets.sdk.st_api.StopPipelineCommand`.
        """
        logger.info('Stopping pipeline %s...', pipeline.id)
        stop_command = self.api_client.stop_pipeline(pipeline_id=pipeline.id)

        if kwargs.get('force', False):
            # Note: Pipeline force stop is applicable only after a pipeline stop.
            stop_command = self.api_client.force_stop_pipeline(pipeline_id=pipeline.id)

        if kwargs.get('wait', True):
            stop_command.wait_for_stopped(kwargs.get('timeout_sec', DEFAULT_STOP_TIMEOUT))

        return stop_command

    @_dump_st_log_on_error
    def run_pipeline_preview(
        self,
        pipeline_id,
        rev=0,
        batches=1,
        batch_size=10,
        skip_targets=True,
        end_stage=None,
        timeout=DEFAULT_PREVIEW_SERVER_TIMEOUT_MS,
        stage_outputs_to_override_json=None,
        **kwargs
    ):
        """Run pipeline preview.

        Args:
            pipeline_id (:obj:`str`): The pipeline instance's ID.
            rev (:obj:`int`, optional): Pipeline revision. Default: ``0``.
            batches (:obj:`int`, optional): Number of batches. Default: ``1``.
            batch_size (:obj:`int`, optional): Batch size. Default: ``10``.
            skip_targets (:obj:`bool`, optional): Skip targets. Default: ``True``.
            end_stage (:obj:`str`, optional): End stage. Default: ``None``.
            timeout (:obj:`int`, optional): Server side preview Timeout in milliseconds. Default: ``120000``.
            stage_outputs_to_override_json (:obj:`str`, optional): Stage outputs to override. Default: ``None``.
            remote (:obj:`bool`, optional): Remote preview (i.e. run on the cluster). Default: ``False``.
            timeout_sec (:obj:`int`, optional): Client side preview timeout, in seconds.
                Default: :py:const:`streamsets.sdk.st.DEFAULT_PREVIEW_CLIENT_TIMEOUT_SEC`.
            wait (:obj:`bool`, optional): Wait for pipeline preview to finish. Default: ``True``.
            time_between_checks (:obj:`int`, optional): Time to sleep between preview status checks. Applicable when
                ```wait``` is enabled, in seconds.
                Default: :py:const:`streamsets.sdk.st.DEFAULT_PREVIEW_TIME_BETWEEN_CHECKS`.

        Returns:
            An instance of :py:class:`streamsets.sdk.st_api.PreviewCommand`.
        """
        logger.info('Running preview for %s...', pipeline_id)
        preview_command = self.api_client.run_pipeline_preview(
            pipeline_id,
            rev,
            batches,
            batch_size,
            skip_targets,
            end_stage,
            timeout,
            stage_outputs_to_override_json,
            kwargs.get('remote', False),
        )
        if kwargs.get('wait', True):
            timeout_sec = kwargs.get('timeout_sec', DEFAULT_PREVIEW_CLIENT_TIMEOUT_SEC)
            time_between_checks = kwargs.get('time_between_checks', DEFAULT_PREVIEW_TIME_BETWEEN_CHECKS)

            if time_between_checks > timeout_sec:
                raise ValueError('Time to sleep between preview status check cannot be greater than preview timeout')

            preview_command.wait_for_finished(timeout_sec=timeout_sec, time_between_checks=time_between_checks)

        return preview_command

    @_dump_st_log_on_error
    def get_pipeline_acl(self, pipeline):
        """Get pipeline ACL.

        Args:
            pipeline (:py:class:`streamsets.sdk.st_models.Pipeline`): The pipeline instance.

        Returns:
            An instance of :py:class:`streamsets.sdk.st_models.PipelineAcl`.
        """
        return st_models.PipelineAcl(self.api_client.get_pipeline_acl(pipeline_id=pipeline.id))

    @_dump_st_log_on_error
    def set_pipeline_acl(self, pipeline, pipeline_acl):
        """Update pipeline ACL.

        Args:
            pipeline (:py:class:`streamsets.sdk.st_models.Pipeline`): The pipeline instance.
            pipeline_acl (:py:class:`streamsets.sdk.st_models.PipelineAcl`): The pipeline ACL instance.

        Returns:
            An instance of :py:class:`streamsets.sdk.st_api.Command`.
        """
        return self.api_client.set_pipeline_acl(pipeline_id=pipeline.id, pipeline_acl_json=pipeline_acl._data)

    @_dump_st_log_on_error
    def get_pipeline_permissions(self, pipeline):
        """Return pipeline permissions for a given pipeline.

        Args:
            pipeline (:obj:`streamsets.sdk.st_models.Pipeline`): The pipeline instance.

        Returns:
            An instance of :py:class:`streamsets.sdk.st_models.PipelinePermissions`.
        """
        return st_models.PipelinePermissions(self.api_client.get_pipeline_permissions(pipeline_id=pipeline.id))

    @_dump_st_log_on_error
    def get_pipeline_status(self, pipeline):
        """Get status of a pipeline.

        Args:
            pipeline (:py:class:`streamsets.sdk.st_models.Pipeline`): The pipeline instance.
        """
        logger.info('Getting status of pipeline %s...', pipeline.id)
        return self.api_client.get_pipeline_status(pipeline_id=pipeline.id)

    @_dump_st_log_on_error
    def get_pipeline_history(self, pipeline):
        """Get a pipeline's history.

        Args:
            pipeline (:py:class:`streamsets.sdk.st_models.Pipeline`): The pipeline instance.

        Returns:
            An instance of :py:class:`streamsets.sdk.st_models.History`.
        """
        logger.info('Getting pipeline history for %s...', pipeline.id)
        return st_models.History(self.api_client.get_pipeline_history(pipeline_id=pipeline.id))

    @_dump_st_log_on_error
    def get_pipeline(self, pipeline_id):
        """Get a pipeline.

        Args:
            pipeline_id (:obj:`str`): Id of pipeline.

        Returns:
            An instance of :py:class:`streamsets.sdk.st_models.Pipeline`.
        """
        exported_pipeline = self.api_client.export_pipeline(pipeline_id=pipeline_id)
        builder = self.get_pipeline_builder()
        builder.import_pipeline(exported_pipeline)
        return builder.build(title=exported_pipeline['pipelineConfig']['title'])

    def get_pipeline_metrics(self, pipeline):
        """Get a pipeline's metrics.

        Args:
            pipeline (:py:class:`streamsets.sdk.st_models.Pipeline`): The pipeline instance.

        Returns:
            An instance of :py:class:`streamsets.sdk.st_models.Metrics`.
        """
        logger.info('Getting metrics for pipeline %s ...', pipeline.id)
        # Mirror ST error handling for pipeline metrics REST API (https://git.io/JfTBI). That is, the API client
        # will handle cases of non-existent pipeline IDs being passed through, but pipelines without metrics
        # (e.g. those are that stopped) will simply return None implicitly.
        command = self.api_client.get_pipeline_metrics(pipeline_id=pipeline.id)
        # Handle case of empty response with HTTP 200 (seen for pipelines while starting) and of HTTP 204.
        if command.response.text and command.response.status_code != requests.codes.no_content:
            return st_models.Metrics(command.response.json())

    def wait_for_pipeline_metric(self, pipeline, metric, value, timeout_sec=DEFAULT_WAIT_FOR_METRIC_TIMEOUT):
        """Block until a pipeline metric reaches the desired value.

        Args:
            pipeline (:py:class:`streamsets.sdk.st_models.Pipeline`): The pipeline instance.
            metric (:py:obj:`str`): The desired metric (e.g. ``'output_record_count'`` or ``'data_batch_count'``).
            value: The desired value to wait for.
            timeout_sec (:obj:`int`, optional): Timeout to wait for ``metric`` to reach ``value``, in seconds.
                Default: :py:const:`streamsets.sdk.st.DEFAULT_WAIT_FOR_METRIC_TIMEOUT`.

        Raises:
            TimeoutError: If ``timeout_sec`` passes without ``metric`` reaching ``value``.
        """
        logger.info('Waiting for pipeline metric %s to reach at value %s ...', metric, value)
        start_waiting_time = time.time()
        stop_waiting_time = start_waiting_time + timeout_sec

        while time.time() < stop_waiting_time:
            metrics = self.get_pipeline_metrics(pipeline)
            if not metrics:
                raise ValueError('Operation only supported for running pipelines')
            current_value = getattr(metrics.pipeline, metric)
            logger.debug('Pipeline metric %s has current value %s ...', metric, current_value)
            if current_value >= value:
                logger.info(
                    'Metric %s reached value %s for pipeline %s (took %.2f s).',
                    metric,
                    value,
                    pipeline.id,
                    time.time() - start_waiting_time,
                )
                break

            time.sleep(1)
        else:
            # We got out of the loop and did not get the metric we were waiting for.
            raise TimeoutError(
                'Metric {} did not reach value {} '
                'after {} s (current value {})'.format(metric, value, timeout_sec, current_value)
            )

    @property
    def transformer_configuration(self):
        """Return all configurations for StreamSets Transformer.
        Returns:
            A :obj:`dict` with property names as keys and property values as values.
        """
        return self.api_client.get_transformer_configuration().response.json()

    @property
    def id(self):
        """Return id for StreamSets Transformer.

        Returns:
            A :obj:`str` Transformer ID.
        """
        transformer_id_command = self.api_client.get_transformer_id()
        transformer_id_json = transformer_id_command.response.json() if transformer_id_command.response.content else {}
        return transformer_id_json['id'] if transformer_id_json else None

    @property
    def current_user(self):
        """Get currently logged-in user and its groups and roles.

        Returns:
            An instance of :py:class:`streamsets.sdk.st_models.User`.
        """
        logger.info('Getting current user ...')
        return st_models.User(self.api_client.get_current_user())

    def change_password(self, old_password, new_password):
        """Change password for the current user.

        Args:
            old_password (:obj:`str`): old password.
            new_password (:obj:`str`): new password.

        Returns:
            An instance of :py:class:`streamsets.sdk.st_api.Command`.
        """
        data = {
            'data': {'id': self.current_user.name, 'oldPassword': old_password, 'newPassword': new_password},
            'envelopeVersion': "1",
        }
        return self.api_client.change_password(data)

    def get_logs(self, ending_offset=-1, extra_message=None, pipeline=None, severity=None):
        """Get logs.

        Args:
            ending_offset (:obj:`int`): ending_offset, Default: ``-1``.
            extra_message (:obj:`str`): extra_message, Default: ``None``.
            pipeline (:py:class:`streamsets.sdk.st_models.Pipeline`): The pipeline instance, Default: ``None``.
            severity (:obj:`str`): severity, Default: ``None``.

        Returns:
            An instance of :py:class:`streamsets.sdk.st_models.Log`.
        """
        pipeline_id = pipeline.id if pipeline is not None else None
        return st_models.Log(self.api_client.get_logs(ending_offset, extra_message, pipeline_id, severity))

    def get_alerts(self):
        """Get pipeline alerts.

        Returns:
            An instance of :py:class:`streamsets.sdk.st_models.Alerts`.
        """
        return st_models.Alerts(self.api_client.get_alerts())

    @property
    def stage_libraries(self):
        """Get all stage libraries.

        Returns:
            A :py:obj:`streamsets.sdk.utils.SeekableList` of
                :py:class:`streamsets.sdk.st_models.StageLibrary` instances.
        """
        stage_libraries = SeekableList()
        repositories = self.api_client.get_stage_libraries_list(repo_url=None, installed_only=False).response.json()
        for repository in repositories:
            repository_manifest = {key: value for key, value in repository.items() if key != 'stageLibraries'}
            for stage_library in repository['stageLibraries']:
                stage_libraries.append(st_models.StageLibrary(stage_library, repository_manifest))
        return stage_libraries

    @property
    def log_configuration(self):
        """Return log configurations for StreamSets Transformer.

        Returns:
            A :obj:`str` log4j properties file contents.
        """
        return self.api_client.get_st_log_config().response.text

    @_dump_st_log_on_error
    def get_bundle_generators(self):
        """Get available support bundle generators.

        Returns:
            An instance of :py:class:`streamsets.sdk.st_models.BundleGenerators`.
        """
        return st_models.BundleGenerators(self.api_client.get_bundle_generators())

    @_dump_st_log_on_error
    def get_bundle(self, generators=None):
        """Generate new support bundle.

        Returns:
            An instance of :py:class:`zipfile.ZipFile`.
        """
        return self.api_client.get_bundle(generators)

    def get_jmx_metrics(self):
        """Get Transformer JMX metrics

        Returns:
            An instance of :py:class:`streamsets.sdk.st_models.JmxMetrics`.
        """
        logger.info('Getting JMX metrics ...')
        return st_models.JmxMetrics(self.api_client.get_jmx_metrics().response.json())



# Copyright 2024 StreamSets Inc.

# fmt: off
import pytest

from streamsets.sdk.sch_models import Permission

# fmt: on


class MockControlHub:
    def __init__(self):
        self.api_client = MockApiClient()


class MockApiClient:
    def update_job_permissions(self, *args, **kwargs):
        return True

    def update_pipeline_permissions(self, *args, **kwargs):
        return True

    def update_sdc_permissions(self, *args, **kwargs):
        return True

    def update_report_definition_permissions(self, *args, **kwargs):
        return True

    def update_connection_permissions(self, *args, **kwargs):
        return True

    def update_topology_permissions(self, *args, **kwargs):
        return True

    def update_subscription_permissions(self, *args, **kwargs):
        return True

    def update_legacy_deployment_permissions(self, *args, **kwargs):
        return True

    def update_deployment_permissions(self, *args, **kwargs):
        return True

    def update_environment_permissions(self, *args, **kwargs):
        return True

    def update_scheduled_task_permissions(self, *args, **kwargs):
        return True

    def update_alert_permissions(self, *args, **kwargs):
        return True


@pytest.fixture(scope="function")
def dummy_permission():
    return Permission(
        {
            'resourceId': '1fd7acec-817f-4ee4-b0cc-7046a06744e7:35be715e-e8e9-11ec-8e84-93331d4150d4',
            'subjectId': '2cfced39-e8e9-11ec-8e84-d7b86be914f5@35be715e-e8e9-11ec-8e84-93331d4150d4',
            'subjectType': 'USER',
            'lastModifiedBy': '2cfced39-e8e9-11ec-8e84-d7b86be914f5@35be715e-e8e9-11ec-8e84-93331d4150d4',
            'lastModifiedOn': 1721317112806,
            'actions': ['READ', 'WRITE'],
        },
        resource_type="CSP_DEPLOYMENT",
        api_client=MockApiClient(),
    )


def test_update_permission_action(dummy_permission):
    valid_permission_actions = [["READ"], ["READ", "WRITE"], ["WRITE", "READ"], ["EXECUTE", "READ"]]
    for action in valid_permission_actions:
        dummy_permission.actions = action
        assert dummy_permission.actions == action


def test_update_permission_action_fails(dummy_permission):
    invalid_permission_actions = [["EAD"], ["WRITE"], ["WRITE", "EXECUTE"], ["EXECUTE", "RAD"]]
    for action in invalid_permission_actions:
        with pytest.raises(ValueError):
            dummy_permission.actions = action
